# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Log to `MLflow <https://www.mlflow.org/docs/latest/index.html>."""

from __future__ import annotations

import fnmatch
import logging
import os
import pathlib
import posixpath
import textwrap
import time
import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Sequence, Union

import numpy as np
import torch

from composer.core.state import State
from composer.loggers.logger import Logger
from composer.loggers.logger_destination import LoggerDestination
from composer.utils import MissingConditionalImportError, dist

if TYPE_CHECKING:
    from mlflow import ModelVersion  # pyright: ignore[reportGeneralTypeIssues]

log = logging.getLogger(__name__)

__all__ = ['MLFlowLogger']

DEFAULT_MLFLOW_EXPERIMENT_NAME = 'my-mlflow-experiment'


class MLFlowLogger(LoggerDestination):
    """Log to `MLflow <https://www.mlflow.org/docs/latest/index.html>`_.

    Args:
        experiment_name: (str, optional): MLflow experiment name. If not set it will be
            use the MLflow environment variable or a default value
        run_name: (str, optional): MLflow run name. If not set it will be the same as the
            Trainer run name
        tags: (dict, optional): MLflow tags to log with the run
        tracking_uri (str | pathlib.Path, optional): MLflow tracking uri, the URI to the
            remote or local endpoint where logs are stored (If none it is set to MLflow default)
        rank_zero_only (bool, optional): Whether to log only on the rank-zero process
            (default: ``True``).
        flush_interval (int): The amount of time, in seconds, that MLflow must wait between
            logging batches of metrics. Any metrics that are recorded by Composer during
            this interval are enqueued, and the queue is flushed when the interval elapses
            (default: ``10``).
        model_registry_prefix (str, optional): The prefix to use when registering models.
            If registering to Unity Catalog, must be in the format ``{catalog_name}.{schema_name}``.
            (default: `''`)
        model_registry_uri (str, optional): The URI of the model registry to use. To register models
            to Unity Catalog, set to ``databricks-uc``. (default: None)
        synchronous (bool, optional): Whether to log synchronously. If ``True``, Mlflow will log
            synchronously to the MLflow backend. If ``False``, Mlflow will log asynchronously. (default: ``False``)
        log_system_metrics (bool, optional): Whether to log system metrics. If ``True``, Mlflow will
            log system metrics (CPU/GPU/memory/network usage) during training. (default: ``True``)
        rename_metrics (dict[str, str], optional): A dict to rename metrics, requires an exact match on the key (default: ``None``)
        ignore_metrics (list[str], optional): A list of glob patterns for metrics to ignore when logging. (default: ``None``)
        ignore_hyperparameters (list[str], optional): A list of glob patterns for hyperparameters to ignore when logging. (default: ``None``)
        run_group (str, optional): A string to group runs together. (default: ``None``)
        resume (bool, optional): If ``True``, Composer will search for an existing run tagged with
            the `run_name` and resume it. If no existing run is found, a new run will be created.
            If ``False``, Composer will create a new run. (default: ``False``)
        logging_buffer_seconds (int, optional): The amount of time, in seconds, that MLflow
            waits before sending logs to the MLflow tracking server. Metrics/params/tags logged
            within this buffer time will be grouped in batches before being sent to the backend.
    """

    def __init__(
        self,
        experiment_name: Optional[str] = None,
        run_name: Optional[str] = None,
        tags: Optional[dict[str, Any]] = None,
        tracking_uri: Optional[Union[str, pathlib.Path]] = None,
        rank_zero_only: bool = True,
        flush_interval: int = 10,
        model_registry_prefix: str = '',
        model_registry_uri: Optional[str] = None,
        synchronous: bool = False,
        log_system_metrics: bool = True,
        rename_metrics: Optional[dict[str, str]] = None,
        ignore_metrics: Optional[list[str]] = None,
        ignore_hyperparameters: Optional[list[str]] = None,
        run_group: Optional[str] = None,
        resume: bool = False,
        logging_buffer_seconds: Optional[int] = 10,
    ) -> None:
        try:
            import mlflow
            from mlflow import MlflowClient
        except ImportError as e:
            raise MissingConditionalImportError(
                extra_deps_group='mlflow',
                conda_package='mlflow',
                conda_channel='conda-forge',
            ) from e
        self._enabled = (not rank_zero_only) or dist.get_global_rank() == 0

        self.experiment_name = experiment_name
        self.run_name = run_name
        self.run_group = run_group
        self.tags = tags or {}
        self.model_registry_prefix = model_registry_prefix
        self.model_registry_uri = model_registry_uri
        self.synchronous = synchronous
        self.log_system_metrics = log_system_metrics
        self.rename_metrics = {} if rename_metrics is None else rename_metrics
        self.ignore_metrics = [] if ignore_metrics is None else ignore_metrics
        self.ignore_hyperparameters = [] if ignore_hyperparameters is None else ignore_hyperparameters
        if self.model_registry_uri == 'databricks-uc':
            if len(self.model_registry_prefix.split('.')) != 2:
                raise ValueError(
                    f'When registering to Unity Catalog, model_registry_prefix must be in the format ' +
                    f'{{catalog_name}}.{{schema_name}}, but got {self.model_registry_prefix}',
                )
        self.resume = resume

        if logging_buffer_seconds:
            os.environ['MLFLOW_ASYNC_LOGGING_BUFFERING_SECONDS'] = str(logging_buffer_seconds)

        self._rank_zero_only = rank_zero_only
        self._last_flush_time = time.time()
        self._flush_interval = flush_interval

        self._experiment_id: Optional[str] = None
        self._run_id = None
        self.run_url = None

        if self._enabled:
            if tracking_uri is None and os.getenv('DATABRICKS_TOKEN') is not None:
                tracking_uri = 'databricks'
            if tracking_uri is None:
                tracking_uri = mlflow.get_tracking_uri()
            self.tracking_uri = str(tracking_uri)
            mlflow.set_tracking_uri(self.tracking_uri)

            if self.model_registry_uri is not None:
                mlflow.set_registry_uri(self.model_registry_uri)
            # Set up MLflow state
            self._run_id = None
            if self.experiment_name is None:
                self.experiment_name = os.getenv(
                    mlflow.environment_variables.MLFLOW_EXPERIMENT_NAME.name,  # type: ignore
                    DEFAULT_MLFLOW_EXPERIMENT_NAME,
                )
            assert self.experiment_name is not None  # type hint

            if os.getenv(
                'DATABRICKS_TOKEN',
            ) is not None and not self.experiment_name.startswith((
                '/Users/',
                '/Shared/',
            )):
                try:
                    from databricks.sdk import WorkspaceClient
                except ImportError as e:
                    raise MissingConditionalImportError(
                        extra_deps_group='mlflow',
                        conda_package='databricks-sdk',
                        conda_channel='conda-forge',
                    ) from e
                databricks_username = WorkspaceClient().current_user.me().user_name or ''
                self.experiment_name = os.path.join('/Users', databricks_username, self.experiment_name.strip('/'))

            self._mlflow_client = MlflowClient(self.tracking_uri)
            # Set experiment
            env_exp_id = os.getenv(
                mlflow.environment_variables.MLFLOW_EXPERIMENT_ID.name,  # pyright: ignore[reportGeneralTypeIssues]
                None,
            )
            if env_exp_id is not None:
                self._experiment_id = env_exp_id
            else:
                exp_from_name = self._mlflow_client.get_experiment_by_name(name=self.experiment_name)
                if exp_from_name is not None:
                    self._experiment_id = exp_from_name.experiment_id
                else:
                    self._experiment_id = (self._mlflow_client.create_experiment(name=self.experiment_name))

    def _start_mlflow_run(self, state):
        import mlflow

        env_run_id = os.getenv(
            mlflow.environment_variables.MLFLOW_RUN_ID.name,  # pyright: ignore[reportGeneralTypeIssues]
            None,
        )
        if env_run_id is not None:
            self._run_id = env_run_id
        elif self.resume:
            # Search for an existing run tagged with this Composer run if `self.resume=True`.
            assert self._experiment_id is not None
            run_name = self.tags['run_name']
            existing_runs = mlflow.search_runs(
                experiment_ids=[self._experiment_id],
                filter_string=f'tags.run_name = "{run_name}"',
                output_format='list',
            )

            if len(existing_runs) > 0:
                self._run_id = existing_runs[0].info.run_id
                log.debug(f'Resuming mlflow run with run id: {self._run_id}')
            else:
                log.debug(
                    'Creating a new mlflow run as `resume` was set to True but no previous run was '
                    'found.',
                )
                new_run = self._mlflow_client.create_run(
                    experiment_id=self._experiment_id,
                    run_name=self.run_name,
                )
                self._run_id = new_run.info.run_id
        else:
            # Create a new run if `env_run_id` is not set or `self.resume=False`.
            new_run = self._mlflow_client.create_run(
                experiment_id=self._experiment_id,
                run_name=self.run_name,
            )
            self._run_id = new_run.info.run_id

        tags = self.tags or {}
        if self.run_group:
            tags['run_group'] = self.run_group
        mlflow.start_run(
            run_id=self._run_id,
            tags=self.tags,
            log_system_metrics=self.log_system_metrics,
        )

    def init(self, state: State, logger: Logger) -> None:
        del logger  # unused

        if self.run_name is None:
            self.run_name = state.run_name

        # Store the Composer run name in the MLFlow run tags so it can be retrieved for autoresume
        self.tags['run_name'] = os.environ.get('RUN_NAME', state.run_name)

        # Adjust name and group based on `rank_zero_only`.
        if not self._rank_zero_only:
            self.run_name += f'-rank{dist.get_global_rank()}'

        # Start run
        if self._enabled:
            self._start_mlflow_run(state)

        # If rank zero only, broadcast the MLFlow experiment and run IDs to other ranks, so the MLFlow run info is
        # available to other ranks during runtime.
        if self._rank_zero_only:
            mlflow_ids_list = [self._experiment_id, self._run_id]
            dist.broadcast_object_list(mlflow_ids_list, src=0)
            self._experiment_id, self._run_id = mlflow_ids_list

    def after_load(self, state: State, logger: Logger) -> None:
        logger.log_hyperparameters({'mlflow_experiment_id': self._experiment_id, 'mlflow_run_id': self._run_id})
        self.run_url = posixpath.join(
            os.environ.get('DATABRICKS_HOST', ''),
            'ml',
            'experiments',
            str(self._experiment_id),
            'runs',
            str(self._run_id),
        )

    def log_table(
        self,
        columns: list[str],
        rows: list[list[Any]],
        name: str = 'Table',
        step: Optional[int] = None,
    ) -> None:
        del step
        if self._enabled:
            try:
                import pandas as pd
            except ImportError as e:
                raise MissingConditionalImportError(
                    extra_deps_group='pandas',
                    conda_package='pandas',
                    conda_channel='conda-forge',
                ) from e
            table = pd.DataFrame.from_records(data=rows, columns=columns)
            assert isinstance(self._run_id, str)
            self._mlflow_client.log_table(
                run_id=self._run_id,
                data=table,
                artifact_file=f'{name}.json',
            )

    def rename(self, key: str):
        return self.rename_metrics.get(key, key)

    def log_metrics(self, metrics: dict[str, Any], step: Optional[int] = None) -> None:
        from mlflow import log_metrics
        if self._enabled:
            # Convert all metrics to floats to placate mlflow.
            metrics = {
                self.rename(k): float(v)
                for k, v in metrics.items()
                if not any(fnmatch.fnmatch(k, pattern) for pattern in self.ignore_metrics)
            }
            log_metrics(
                metrics=metrics,
                step=step,
                synchronous=self.synchronous,
            )

    def log_hyperparameters(self, hyperparameters: dict[str, Any]):
        from mlflow import log_params

        if self._enabled:
            hyperparameters = {
                k: v
                for k, v in hyperparameters.items()
                if not any(fnmatch.fnmatch(k, pattern) for pattern in self.ignore_hyperparameters)
            }
            log_params(
                params=hyperparameters,
                synchronous=self.synchronous,
            )

    def register_model(
        self,
        model_uri: str,
        name: str,
        await_registration_for: int = 300,
        tags: Optional[dict[str, Any]] = None,
    ) -> 'ModelVersion':
        """Register a model to model registry.

        Args:
            model_uri (str): The URI of the model to register.
            name (str): The name of the model to register. Will be appended to ``model_registry_prefix``.
            await_registration_for (int, optional): The number of seconds to wait for the model to be registered.
                Defaults to 300.
            tags (Optional[dict[str, Any]], optional): A dictionary of tags to add to the model. Defaults to None.
            registry_uri (str, optional): The URI of the model registry. Defaults to `None` which will register to
                the Databricks Unity Catalog.

        Returns:
            ModelVersion: The registered model.
        """
        if self._enabled:
            full_name = f'{self.model_registry_prefix}.{name}' if len(self.model_registry_prefix) > 0 else name

            import mlflow
            return mlflow.register_model(
                model_uri=model_uri,
                name=full_name,
                await_registration_for=await_registration_for,
                tags=tags,
            )

    def save_model(self, flavor: Literal['transformers', 'peft'], **kwargs):
        """Save a model to MLflow.

        Note: The ``'peft'`` flavor is experimental and the API is subject to change without warning.

        Args:
            flavor (Literal['transformers', 'peft']): The MLflow model flavor to use. Currently only ``'transformers'`` and ``'peft'`` are supported.
            **kwargs: Keyword arguments to pass to the MLflow model saving function.

        Raises:
            NotImplementedError: If ``flavor`` is not ``'transformers'`` or ``'peft'``.
        """
        if self._enabled:
            import mlflow
            if flavor == 'transformers':
                mlflow.transformers.save_model(**kwargs)
            elif flavor == 'peft':
                import transformers

                # TODO: Remove after mlflow fixes the bug that makes this necessary
                mlflow.store._unity_catalog.registry.rest_store.get_feature_dependencies = lambda *args, **kwargs: ''  # type: ignore

                # This is a temporary workaround until MLflow adds full support for saving PEFT models.
                # https://github.com/mlflow/mlflow/issues/9256
                log.warning(
                    'Saving PEFT models using MLflow is experimental and the API is subject to change without warning.',
                )
                expected_keys = {'path', 'save_pretrained_dir'}
                if not expected_keys.issubset(kwargs.keys()):
                    raise ValueError(f'Expected keys {expected_keys} but got {kwargs.keys()}')

                # This does not implement predict for now, as we will wait for the full MLflow support
                # for PEFT models.
                class PeftModel(mlflow.pyfunc.PythonModel):

                    def load_context(self, context):
                        self.model = transformers.AutoModelForCausalLM.from_pretrained(
                            context.artifacts['lora_checkpoint'],
                        )
                        self.tokenizer = transformers.AutoTokenizer.from_pretrained(
                            context.artifacts['lora_checkpoint'],
                        )

                from mlflow.models.signature import ModelSignature
                from mlflow.types import ColSpec, DataType, Schema

                # This is faked for now, until MLflow adds full support for saving PEFT models.
                input_schema = Schema([
                    ColSpec(DataType.string, 'fake_input'),
                ])
                output_schema = Schema([ColSpec(DataType.string)])
                signature = ModelSignature(inputs=input_schema, outputs=output_schema)

                # Symlink the directory so that we control the path that MLflow saves the model under
                os.symlink(kwargs['save_pretrained_dir'], 'lora_checkpoint')

                mlflow.pyfunc.save_model(
                    path=kwargs['path'],
                    artifacts={'lora_checkpoint': 'lora_checkpoint'},
                    python_model=PeftModel(),
                    signature=signature,
                )

                os.unlink('lora_checkpoint')
            else:
                raise NotImplementedError(f'flavor {flavor} not supported.')

    def log_model(self, flavor: Literal['transformers'], **kwargs):
        """Log a model to MLflow.

        Args:
            flavor (Literal['transformers']): The MLflow model flavor to use. Currently only ``'transformers'`` is supported.
            **kwargs: Keyword arguments to pass to the MLflow model logging function.

        Raises:
            NotImplementedError: If ``flavor`` is not ``'transformers'``.
        """
        if self._enabled:
            import mlflow
            if flavor == 'transformers':
                mlflow.transformers.log_model(**kwargs)
            else:
                raise NotImplementedError(f'flavor {flavor} not supported.')

    def register_model_with_run_id(
        self,
        model_uri: str,
        name: str,
        await_creation_for: int = 300,
        tags: Optional[dict[str, Any]] = None,
    ):
        """Similar to ``register_model``, but uses a different MLflow API to allow passing in the run id.

        Args:
            model_uri (str): The URI of the model to register.
            name (str): The name of the model to register. Will be appended to ``model_registry_prefix``.
            await_creation_for (int, optional): The number of seconds to wait for the model to be registered. Defaults to 300.
            tags (Optional[dict[str, Any]], optional): A dictionary of tags to add to the model. Defaults to None.
        """
        if self._enabled:
            from mlflow.exceptions import MlflowException
            from mlflow.protos.databricks_pb2 import ALREADY_EXISTS, RESOURCE_ALREADY_EXISTS, ErrorCode

            full_name = f'{self.model_registry_prefix}.{name}' if len(self.model_registry_prefix) > 0 else name

            # This try/catch code is copied from
            # https://github.com/mlflow/mlflow/blob/3ba1e50e90a38be19920cb9118593a43d7cfa90e/mlflow/tracking/_model_registry/fluent.py#L90-L103
            try:
                create_model_response = self._mlflow_client.create_registered_model(full_name)
                log.info(f'Successfully registered model {name} with {create_model_response.name}')
            except MlflowException as e:
                if e.error_code in (
                    ErrorCode.Name(RESOURCE_ALREADY_EXISTS),
                    ErrorCode.Name(ALREADY_EXISTS),
                ):
                    log.info(f'Registered model {name} already exists. Creating a new version of this model...')
                else:
                    raise e

            create_version_response = self._mlflow_client.create_model_version(
                name=full_name,
                source=model_uri,
                run_id=self._run_id,
                await_creation_for=await_creation_for,
                tags=tags,
            )

            log.info(
                f'Successfully created model version {create_version_response.version} for model {create_version_response.name}',
            )

    def log_images(
        self,
        images: Union[np.ndarray, torch.Tensor, Sequence[Union[np.ndarray, torch.Tensor]]],
        name: str = 'image',
        channels_last: bool = False,
        step: Optional[int] = None,
        masks: Optional[dict[str, Union[np.ndarray, torch.Tensor, Sequence[Union[np.ndarray, torch.Tensor]]]]] = None,
        mask_class_labels: Optional[dict[int, str]] = None,
        use_table: bool = True,
    ):
        unused_args = (masks, mask_class_labels)  # Unused (only for wandb)
        if any(unused_args):
            warnings.warn(
                textwrap.dedent(
                    f"""MLFlowLogger does not support masks, class labels, or tables of images,
                          but got masks={masks}, mask_class_labels={mask_class_labels}""",
                ),
            )
        if self._enabled:
            if not isinstance(images, Sequence) and images.ndim <= 3:
                images = [images]
            for im_ind, image in enumerate(images):
                image = _convert_to_mlflow_image(image, channels_last)
                assert isinstance(self._run_id, str)
                self._mlflow_client.log_image(
                    image=image,
                    key=f'{name}_{step}_{im_ind}',
                    run_id=self._run_id,
                    step=step,
                )

    def post_close(self):
        if self._enabled:
            import mlflow

            assert isinstance(self._run_id, str)
            mlflow.flush_async_logging()
            self._mlflow_client.set_terminated(self._run_id)
            mlflow.end_run()


def _convert_to_mlflow_image(image: Union[np.ndarray, torch.Tensor], channels_last: bool) -> np.ndarray:
    if isinstance(image, torch.Tensor):
        image = image.data.cpu().numpy()

    # Error out for empty arrays or weird arrays of dimension 0.
    if np.any(np.equal(image.shape, 0)):
        raise ValueError(f'Got an image (shape {image.shape}) with at least one dimension being 0! ')

    # Squeeze any singleton dimensions and then add them back in if image dimension
    # less than 3.
    image = image.squeeze()

    # Add in length-one dimensions to get back up to 3
    # putting channels last.
    if image.ndim == 1:
        image = np.expand_dims(image, (1, 2))
        channels_last = True
    if image.ndim == 2:
        image = np.expand_dims(image, 2)
        channels_last = True

    if image.ndim != 3:
        raise ValueError(
            textwrap.dedent(
                f'''Input image must be 3 dimensions, but instead
                            got {image.ndim} dims at shape: {image.shape}
                            Your input image was interpreted as a batch of {image.ndim}
                            -dimensional images because you either specified a
                            {image.ndim + 1}D image or a list of {image.ndim}D images.
                            Please specify either a 4D image of a list of 3D images''',
            ),
        )

    assert isinstance(image, np.ndarray)
    if not channels_last:
        image = image.transpose(1, 2, 0)
    if image.shape[-1] not in [1, 3, 4]:
        raise ValueError(
            textwrap.dedent(
                f'''Input image must have 1, 3, or 4 channels, but instead
                            got {image.shape[-1]} channels at shape: {image.shape}
                            Please specify either a 1-, 3-, or 4-channel image or a list of
                            1-, 3-, or 4-channel images''',
            ),
        )
    return image



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""The ComposerModel base interface."""
from __future__ import annotations

import abc
import copy
from typing import Any, Optional, Sequence, Union

import torch
from torch import Tensor
from torchmetrics import Metric

from composer.core import Batch
from composer.loggers import Logger

__all__ = ['ComposerModel']


class ComposerModel(torch.nn.Module, abc.ABC):
    """The interface needed to make a PyTorch model compatible with :class:`composer.Trainer`.

    To create a :class:`.Trainer`\\-compatible model, subclass :class:`.ComposerModel` and
    implement :meth:`forward` and :meth:`loss`. For full functionality (logging and validation), implement :meth:`get_metrics`
    and :meth:`eval_forward`.

    See the :doc:`Composer Model walk through </composer_model>` for more details.

    Minimal Example:

    .. code-block:: python

        import torchvision
        import torch.nn.functional as F

        from composer.models import ComposerModel

        class ResNet18(ComposerModel):

            def __init__(self):
                super().__init__()
                self.model = torchvision.models.resnet18() # define PyTorch model in __init__.

            def forward(self, batch): # batch is the output of the dataloader
                # specify how batches are passed through the model
                inputs, _ = batch
                return self.model(inputs)

            def loss(self, outputs, batch):
                # pass batches and `forward` outputs to the loss
                _, targets = batch
                return F.cross_entropy(outputs, targets)

    Attributes:
        logger (Optional[Logger]): The training :class:`.Logger`.
            The trainer sets the :class:`.Logger` on the:attr:`.Event.INIT` event.
    """

    def __init__(self) -> None:
        super().__init__()
        self.logger: Optional[Logger] = None

    def __deepcopy__(self, memo: dict):
        # From https://stackoverflow.com/questions/1500718/how-to-override-the-copy-deepcopy-operations-for-a-python-object
        # The `logger` should not be copied
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == 'logger':
                copied_v = v
            else:
                copied_v = copy.deepcopy(v, memo)
            setattr(result, k, copied_v)
        return result

    def __copy__(self):
        # From https://stackoverflow.com/questions/1500718/how-to-override-the-copy-deepcopy-operations-for-a-python-object
        # Need to manually define `__copy__` so it does not rely on `__getstate__`, which would not copy the logger.
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __getstate__(self):
        # Don't pickle the logger
        state = self.__dict__.copy()
        state['logger'] = None
        return state

    @abc.abstractmethod
    def forward(self, batch: Batch) -> Any:
        """Compute model output given a batch from the dataloader.

        Args:
            batch (~composer.core.types.Batch): The output batch from dataloader.

        Returns:
            Any:
                The result that is passed to :meth:`loss` as the parameter :attr:`outputs`.

        .. warning:: This method is different from vanilla PyTorch ``model.forward(x)`` or ``model(x)`` as it takes a
                     batch of data that has to be unpacked.

        Example:

        .. code-block:: python

            def forward(self, batch): # batch is the output of the dataloader
                inputs, _ = batch
                return self.model(inputs)

        The outputs of :meth:`forward` are passed to :meth:`loss` by the trainer:

        .. code-block:: python

            for batch in train_dataloader:
                optimizer.zero_grad()
                outputs = model.forward(batch)
                loss = model.loss(outputs, batch)
                loss.backward()
        """
        pass

    @abc.abstractmethod
    def loss(self, outputs: Any, batch: Batch, *args, **kwargs) -> Union[Tensor, Sequence[Tensor]]:
        """Compute the loss of the model given ``outputs`` from :meth:`forward` and a
        :class:`~composer.core.types.Batch` of data from the dataloader. The :class:`.Trainer`
        will call ``.backward()`` on the returned loss.

        Args:
            outputs (Any): The output of the forward pass.
            batch (~composer.core.types.Batch): The output batch from dataloader.

        Returns:
            Tensor | Sequence[Tensor]: The loss as a :class:`torch.Tensor`.

        Example:

        .. code-block:: python

            import torch.nn.functional as F

            def loss(self, outputs, batch):
                # pass batches and :meth:`forward` outputs to the loss
                 _, targets = batch # discard inputs from batch
                return F.cross_entropy(outputs, targets)

        The outputs of :meth:`forward` are passed to :meth:`loss` by the trainer:

        .. code-block:: python

            for batch in train_dataloader:
                optimizer.zero_grad()
                outputs = model.forward(batch)
                loss = model.loss(outputs, batch)
                loss.backward()
        """
        pass

    def eval_forward(
        self,
        batch: Any,
        outputs: Optional[Any] = None,
    ) -> Any:
        """Run the evaluation forward pass.

		By default, it returns the ``outputs`` if they are not None. Otherwise,
		``self(batch)`` is returned.

		Override this method for models that require custom validation logic -- e.g. self-supervised learning.

		Args:
			batch: The dataloader batch.
			outputs (Any, optional): If training, the outputs from the forward pass. Otherwise, None.

		Returns:
			Any: The evaluation outputs.
		"""
        return outputs if outputs is not None else self.forward(batch)

    def update_metric(
        self,
        batch: Any,
        outputs: Any,
        metric: Metric,
    ) -> Optional[dict]:
        """Update the given metric.

		Args:
			batch: The dataloader batch
			outputs: The output from :meth:`eval_forward`
			metric (Metric): The metric to update.

        Returns:
            Optional[dict]: Optionally return metric results to be stored in state.
		"""
        raise NotImplementedError()

    def get_metrics(self, is_train: bool) -> dict[str, Metric]:
        """Get the metrics.

        This method will be called by the trainer immediately after :attr:`.Event.INIT`.

        .. note::

            Each item in the returned dictionary will be ``copy.deepcopy`` before it is used. This
            is to ensure that each dataloader (e.g. train, eval) will be accumulating metrics separately.

            To share a metric across all dataloaders, wrap it with ``MetricSpec(metric=metric, share=False)``.

        Args:
            is_train (bool): Whether the training metrics or evaluation metrics should be returned.

        Returns:
            dict[str, Metric]: A mapping of the metric name to a Metric.
        """
        return {}



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Monitor gradients during training."""

import torch

from composer.core import Callback, State
from composer.loggers import Logger
from composer.utils import dist

__all__ = ['OptimizerMonitor']


class OptimizerMonitor(Callback):
    """Computes and logs the L2 norm of gradients as well as any optimizer-specific metrics implemented in the optimizer's `report_per_parameter_metrics` method.

    L2 norms are calculated after the reduction of gradients across GPUs. This function iterates over the parameters of
    the model and may cause a reduction in throughput while training large models. In order to ensure the
    correctness of the norm, this function should be called after gradient unscaling in cases where gradients are scaled.

    Example:
        .. doctest::

            >>> from composer import Trainer
            >>> from composer.callbacks import OptimizerMonitor
            >>> # constructing trainer object with this callback
            >>> trainer = Trainer(
            ...     model=model,
            ...     train_dataloader=train_dataloader,
            ...     eval_dataloader=eval_dataloader,
            ...     optimizers=optimizer,
            ...     max_duration="1ep",
            ...     callbacks=[OptimizerMonitor()],
            ... )

    The metrics are logged by the :class:`.Logger` to the following keys as described below. `grad_l2_norm` and `layer_grad_l2_norm` are
    logged in addition to metrics logged by the optimizer's `report_per_parameter_metrics` method. For convenience we have listed
    the metrics logged by DecoupledAdamW below.

    +-----------------------------------------------+-----------------------------------------------------+
    | Key                                           | Logged data                                         |
    +===============================================+=====================================================+
    |                                               | L2 norm of the gradients of all parameters in       |
    | ``l2_norm/grad/global``                       | the model on the :attr:`.Event.AFTER_TRAIN_BATCH`   |
    |                                               | event.                                              |
    +-----------------------------------------------+-----------------------------------------------------+
    |                                               | Layer-wise L2 norms                                 |
    | ``l2_norm/grad/LAYER_NAME``                   |                                                     |
    |                                               |                                                     |
    +-----------------------------------------------+-----------------------------------------------------+
    |                                               | Layer-wise L2 norms of Adam first moment after      |
    | ``l2_norm/moment/LAYER_NAME``                 |  calling optimizer step.                            |
    |                                               |                                                     |
    +-----------------------------------------------+-----------------------------------------------------+
    |                                               | Layer-wise L2 norms of parameter weights            |
    | ``l2_norm/param/LAYER_NAME``                  |                                                     |
    |                                               |                                                     |
    +-----------------------------------------------+-----------------------------------------------------+
    |                                               | Layer-wise L2 norms of the step                     |
    | ``l2_norm/update/LAYER_NAME``                 |                                                     |
    |                                               |                                                     |
    +-----------------------------------------------+-----------------------------------------------------+
    """

    def __init__(self, log_optimizer_metrics: bool = True, batch_log_interval: int = 10):
        self.log_optimizer_metrics = log_optimizer_metrics
        self.batch_log_interval = batch_log_interval

    def batch_end(self, state: State, logger: Logger):
        if state.timestamp.batch.value % self.batch_log_interval != 0:
            return

        norm = 0.0
        optimizer_metrics = {}

        for name, p in state.model.named_parameters():
            if p.grad is not None and p.requires_grad:

                metric_reporter = getattr(state.optimizers[0], 'report_per_parameter_metrics', None)
                if callable(metric_reporter) and self.log_optimizer_metrics:
                    optimizer_metrics.update(metric_reporter(p, name, optimizer_metrics))

                # Always log grad norm as a default metric if it's not specified
                if f'l2_norm/grad/{name}' not in optimizer_metrics:
                    param_grad_norm = torch.linalg.vector_norm(p.grad)
                    optimizer_metrics[f'l2_norm/grad/{name}'] = param_grad_norm

        if state.fsdp_enabled and dist.get_world_size() > 0 and self.log_optimizer_metrics:
            # If FSDP is enabled, the optimizer state lives on different ranks and must be reduced
            # and combined before we can compute metrics.
            # Each metric has a different way of being reduced, so the optimizer is responsible for implementing
            # the reduction process.
            # It occurs first via a pre-reduce, where the metric on each rank is modified and prepared
            # then an all-reduce where the modified metric on each rank is combined into the correct metric across all ranks.
            #
            # For example, L2 norms are squared on each rank before we apply all_reduce(SUM) and take the sqrt on each rank
            pre_reduce_metrics = getattr(state.optimizers[0], 'pre_reduce_metrics', None)
            if callable(pre_reduce_metrics) and self.log_optimizer_metrics:
                optimizer_metrics = pre_reduce_metrics(optimizer_metrics)

            dist_reduce_metrics = getattr(state.optimizers[0], 'dist_reduce_metrics', None)
            if callable(dist_reduce_metrics) and self.log_optimizer_metrics:
                optimizer_metrics = dist_reduce_metrics(optimizer_metrics)

        for metric in optimizer_metrics:
            if metric.startswith('l2_norm/grad'):
                norm += optimizer_metrics[metric]**2

        optimizer_metrics['l2_norm/grad/global'] = norm**0.5

        for metric in optimizer_metrics:
            if isinstance(optimizer_metrics[metric], torch.Tensor):
                optimizer_metrics[metric] = optimizer_metrics[metric].item()
        logger.log_metrics(optimizer_metrics)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Adds Squeeze-and-Excitation blocks (`Hu et al, 2019 <https://arxiv.org/abs/1709.01507>`_) after the
:class:`~torch.nn.Conv2d` modules in a neural network.

See :class:`~composer.algorithms.SqueezeExcite` or the :doc:`Method Card </method_cards/squeeze_excite>` for details.
"""

from composer.algorithms.squeeze_excite.squeeze_excite import SqueezeExcite as SqueezeExcite
from composer.algorithms.squeeze_excite.squeeze_excite import SqueezeExcite2d as SqueezeExcite2d
from composer.algorithms.squeeze_excite.squeeze_excite import SqueezeExciteConv2d as SqueezeExciteConv2d
from composer.algorithms.squeeze_excite.squeeze_excite import apply_squeeze_excite as apply_squeeze_excite

__all__ = ['SqueezeExcite', 'SqueezeExcite2d', 'SqueezeExciteConv2d', 'apply_squeeze_excite']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import subprocess

import pytest

import composer


@pytest.mark.parametrize(
    'args',
    [
        ['composer', '--version'],
        ['python', '-m', 'composer', '--version'],
        ['python', '-m', 'composer.cli', '--version'],
        ['python', '-m', 'composer.cli.launcher', '--version'],
    ],
)
def test_cli_version(args: list[str]):
    version_str = subprocess.check_output(args, text=True)
    assert version_str == f'MosaicML Composer {composer.__version__}\n'



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""MLflow Artifacts object store."""

from __future__ import annotations

import logging
import os
import pathlib
import tempfile
from typing import Callable, Optional, Union

from composer.utils.import_helpers import MissingConditionalImportError
from composer.utils.object_store.object_store import ObjectStore, ObjectStoreTransientError

__all__ = ['MLFlowObjectStore']

MLFLOW_DATABRICKS_TRACKING_URI = 'databricks'
MLFLOW_DBFS_PATH_PREFIX = 'databricks/mlflow-tracking/'

DEFAULT_MLFLOW_EXPERIMENT_NAME = 'mlflow-object-store'

MLFLOW_EXPERIMENT_ID_FORMAT_KEY = 'mlflow_experiment_id'
MLFLOW_RUN_ID_FORMAT_KEY = 'mlflow_run_id'

MLFLOW_EXPERIMENT_ID_PLACEHOLDER = '{' + MLFLOW_EXPERIMENT_ID_FORMAT_KEY + '}'
MLFLOW_RUN_ID_PLACEHOLDER = '{' + MLFLOW_RUN_ID_FORMAT_KEY + '}'

log = logging.getLogger(__name__)


def _wrap_mlflow_exceptions(uri: str, e: Exception):
    """Wraps retryable MLflow errors in ObjectStoreTransientError for automatic retry handling."""
    from mlflow.exceptions import (
        ABORTED,
        DATA_LOSS,
        DEADLINE_EXCEEDED,
        ENDPOINT_NOT_FOUND,
        INTERNAL_ERROR,
        INVALID_STATE,
        NOT_FOUND,
        REQUEST_LIMIT_EXCEEDED,
        RESOURCE_DOES_NOT_EXIST,
        RESOURCE_EXHAUSTED,
        TEMPORARILY_UNAVAILABLE,
        ErrorCode,
        MlflowException,
    )

    # https://github.com/mlflow/mlflow/blob/39b76b5b05407af5d223e892b03e450b7264576a/mlflow/exceptions.py for used error codes.
    # https://github.com/mlflow/mlflow/blob/39b76b5b05407af5d223e892b03e450b7264576a/mlflow/protos/databricks.proto for code descriptions.
    retryable_server_codes = [
        ErrorCode.Name(code) for code in [
            DATA_LOSS,
            INTERNAL_ERROR,
            INVALID_STATE,
            TEMPORARILY_UNAVAILABLE,
            DEADLINE_EXCEEDED,
        ]
    ]
    retryable_client_codes = [ErrorCode.Name(code) for code in [ABORTED, REQUEST_LIMIT_EXCEEDED, RESOURCE_EXHAUSTED]]
    not_found_codes = [ErrorCode.Name(code) for code in [RESOURCE_DOES_NOT_EXIST, NOT_FOUND, ENDPOINT_NOT_FOUND]]

    if isinstance(e, MlflowException):
        error_code = e.error_code  # pyright: ignore
        if error_code in retryable_server_codes or error_code in retryable_client_codes:
            raise ObjectStoreTransientError(error_code) from e
        elif error_code in not_found_codes:
            raise FileNotFoundError(f'Object {uri} not found') from e

    raise e


def _get_timeout_and_set_socket_default() -> Optional[int]:
    timeout = os.environ.get('MLFLOW_PATCHED_FILE_UPLOAD_TIMEOUT', None)
    if timeout is not None:
        import socket
        timeout = int(timeout)
        socket.setdefaulttimeout(timeout)
    return timeout


# Original source: https://github.com/mlflow/mlflow/blob/a85081631eb665fa25046cb0b7daf0fbbdd5949f/mlflow/azure/client.py#L42
def _patch_adls_file_upload_with_timeout(sas_url, local_file, start_byte, size, position, headers, is_single):
    """Performs an ADLS Azure file create `Patch` operation.

    (https://docs.microsoft.com/en-us/rest/api/storageservices/datalakestoragegen2/path/update)

    Args:
        sas_url: A shared access signature URL referring to the Azure ADLS server
            to which the file update command should be issued.
        local_file: The local file to upload
        start_byte: The starting byte of the local file to upload
        size: The number of bytes to upload
        position: Positional offset of the data in the Patch request
        headers: Additional headers to include in the Patch request body
        is_single: Whether this is the only patch operation for this file
    """
    from mlflow.azure.client import _append_query_parameters, _is_valid_adls_patch_header, _logger
    from mlflow.utils import rest_utils
    from mlflow.utils.file_utils import read_chunk

    new_params = {'action': 'append', 'position': str(position)}
    if is_single:
        new_params['flush'] = 'true'
    request_url = _append_query_parameters(sas_url, new_params)

    request_headers = {}
    for name, value in headers.items():
        if _is_valid_adls_patch_header(name):
            request_headers[name] = value
        else:
            _logger.debug("Removed unsupported '%s' header for ADLS Gen2 Patch operation", name)

    data = read_chunk(local_file, size, start_byte)

    ### Changed here to pass a timeout along to cloud_storage_http_request
    ### And to set the socket timeout
    timeout = _get_timeout_and_set_socket_default()
    with rest_utils.cloud_storage_http_request(
        'patch',
        request_url,
        data=data,
        headers=request_headers,
        timeout=timeout,
    ) as response:
        rest_utils.augmented_raise_for_status(response)
        response.close()


def _put_adls_file_creation_with_timeout(sas_url, headers):
    """Performs an ADLS Azure file create `Put` operation.

    (https://docs.microsoft.com/en-us/rest/api/storageservices/datalakestoragegen2/path/create)

    :param sas_url: A shared access signature URL referring to the Azure ADLS server
                    to which the file creation command should be issued.
    :param headers: Additional headers to include in the Put request body
    """
    from mlflow.azure.client import _append_query_parameters, _is_valid_adls_put_header, _logger
    from mlflow.utils import rest_utils

    request_url = _append_query_parameters(sas_url, {'resource': 'file'})

    request_headers = {}
    for name, value in headers.items():
        if _is_valid_adls_put_header(name):
            request_headers[name] = value
        else:
            _logger.debug("Removed unsupported '%s' header for ADLS Gen2 Put operation", name)

    ### Changed here to pass a timeout along to cloud_storage_http_request
    ### And to set the socket timeout
    timeout = _get_timeout_and_set_socket_default()
    with rest_utils.cloud_storage_http_request(
        'put',
        request_url,
        headers=request_headers,
        timeout=timeout,
    ) as response:
        rest_utils.augmented_raise_for_status(response)
        response.close()


class MLFlowObjectStore(ObjectStore):
    """Utility class for uploading and downloading artifacts from MLflow.

    It can be initialized for an existing run, a new run in an existing experiment, the active run used by the `mlflow`
    module, or a new run in a new experiment. See the documentation for ``path`` for more details.

    .. note::
        At this time, only Databricks-managed MLflow with a 'databricks' tracking URI is supported.
        Using this object store requires configuring Databricks authentication through a configuration file or
        environment variables. See
        https://databricks-sdk-py.readthedocs.io/en/latest/authentication.html#databricks-native-authentication

        Unlike other object stores, the DBFS URI scheme for MLflow artifacts has no bucket, and the path is prefixed
        with the artifacts root directory for a given experiment/run,
        `databricks/mlflow-tracking/<experiment_id>/<run_id>/`. However, object names are also sometimes passed by
        upstream code as artifact paths relative to this root, rather than the full path. To keep upstream code simple,
        :class:`MLFlowObjectStore` accepts both relative MLflow artifact paths and absolute DBFS paths as object names.
        If an object name takes the form of
        `databricks/mlflow-tracking/<experiment_id>/<run_id>/artifacts/<artifact_path>`,
        it is assumed to be an absolute DBFS path, and the `<artifact_path>` is used when uploading objects to MLflow.
        Otherwise, the object name is assumed to be a relative MLflow artifact path, and the full provided name will be
        used as the artifact path when uploading to MLflow.

    Args:
        path (str): A DBFS path of the form
            `databricks/mlflow-tracking/<experiment_id>/<run_id>/artifacts/<path>`.
            `experiment_id` and `run_id` can be set as the format string placeholders
            `{mlflow_experiment_id}` and `{mlflow_run_id}'`.

            If both `experiment_id` and `run_id` are set as placeholders, the MLFlowObjectStore will be associated with
            the currently active MLflow run if one exists. If no active run exists, a new run will be created under a
            default experiment name, or the experiment name specified by the `MLFLOW_EXPERIMENT_NAME` environment
            variable if one is set.

            If `experiment_id` is provided and `run_id` is not, the MLFlowObjectStore will create a new run in the
            provided experiment.

            Providing a `run_id` without an `experiment_id` will raise an error.
        multipart_upload_chunk_size(int, optional): The maximum size of a single chunk in an MLflow multipart upload.
            The maximum number of chunks supported by MLflow is 10,000, so the max file size that can
            be uploaded is `10 000 * multipart_upload_chunk_size`. Defaults to 100MB for a max upload size of 1TB.
    """

    def __init__(self, path: str, multipart_upload_chunk_size: int = 100 * 1024 * 1024) -> None:
        try:
            import mlflow
            from mlflow import MlflowClient
        except ImportError as e:
            raise MissingConditionalImportError('mlflow', conda_package='mlflow>=2.9.2,<3.0') from e

        try:
            from databricks.sdk import WorkspaceClient
        except ImportError as e:
            raise MissingConditionalImportError('databricks', conda_package='databricks-sdk>=0.15.0,<1.0') from e

        # This is a temporary workaround for an intermittent hang we have encountered when uploading files to ADLS.
        # MLflow is working on an upstream fix, but in the meantime, patching in timeouts works around the hang.
        log.debug('Patching MLflow Azure client to include timeout in ADLS file upload')
        mlflow.store.artifact.databricks_artifact_repo.patch_adls_file_upload = _patch_adls_file_upload_with_timeout  # type: ignore
        mlflow.store.artifact.databricks_artifact_repo.put_adls_file_creation = _put_adls_file_creation_with_timeout  # type: ignore

        tracking_uri = os.getenv(
            mlflow.environment_variables.MLFLOW_TRACKING_URI.name,  # pyright: ignore[reportGeneralTypeIssues]
            MLFLOW_DATABRICKS_TRACKING_URI,
        )
        if tracking_uri != MLFLOW_DATABRICKS_TRACKING_URI:
            raise ValueError(
                'MLFlowObjectStore currently only supports Databricks-hosted MLflow tracking. '
                f'Environment variable `MLFLOW_TRACKING_URI` is set to a non-Databricks URI {tracking_uri}. '
                f'Please unset it or set the value to `{MLFLOW_DATABRICKS_TRACKING_URI}`.',
            )

        # Use the Databricks WorkspaceClient to check that credentials are set up correctly.
        try:
            WorkspaceClient()
        except Exception as e:
            raise ValueError(
                f'Databricks SDK credentials not correctly setup. '
                'Visit https://databricks-sdk-py.readthedocs.io/en/latest/authentication.html#databricks-native-authentication '
                'to identify different ways to setup credentials.',
            ) from e

        self._mlflow_client = MlflowClient(tracking_uri)
        mlflow.environment_variables.MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE.set(  # pyright: ignore[reportGeneralTypeIssues]
            multipart_upload_chunk_size,
        )

        experiment_id, run_id, _ = MLFlowObjectStore.parse_dbfs_path(path)
        if experiment_id == MLFLOW_EXPERIMENT_ID_PLACEHOLDER:
            experiment_id = None
        if run_id == MLFLOW_RUN_ID_PLACEHOLDER:
            run_id = None

        # Construct the `experiment_id` and `run_id` depending on whether format placeholders were provided.
        self.experiment_id, self.run_id = self._init_run_info(experiment_id, run_id)

    def _init_run_info(self, experiment_id: Optional[str], run_id: Optional[str]) -> tuple[str, str]:
        """Returns the experiment ID and run ID for the MLflow run backing this object store.

        In a distributed setting, this should only be called on the rank 0 process.
        """
        import mlflow

        if experiment_id is None:
            if run_id is not None:
                raise ValueError('A `run_id` cannot be provided without a valid `experiment_id`.')

            active_run = mlflow.active_run()
            if active_run is not None:
                experiment_id = active_run.info.experiment_id
                run_id = active_run.info.run_id
                log.debug(f'MLFlowObjectStore using active MLflow run {run_id=}')
            else:
                # If no active run exists, create a new run for the default experiment.
                mlflow_env_var_name = mlflow.environment_variables.MLFLOW_EXPERIMENT_NAME.name  # pyright: ignore[reportGeneralTypeIssues]
                experiment_name = os.getenv(mlflow_env_var_name, DEFAULT_MLFLOW_EXPERIMENT_NAME)

                experiment = self._mlflow_client.get_experiment_by_name(experiment_name)
                if experiment is not None:
                    experiment_id = experiment.experiment_id
                else:
                    experiment_id = self._mlflow_client.create_experiment(experiment_name)

                run_id = self._mlflow_client.create_run(experiment_id).info.run_id

                log.debug(
                    f'MLFlowObjectStore using a new MLflow run {run_id=}'
                    f'for new experiment "{experiment_name}" {experiment_id=}',
                )
        else:
            if run_id is not None:
                # If a `run_id` is provided, check that it belongs to the provided experiment.
                run = self._mlflow_client.get_run(run_id)
                if run.info.experiment_id != experiment_id:
                    raise ValueError(
                        f'Provided `run_id` {run_id} does not belong to provided experiment {experiment_id}. '
                        f'Found experiment {run.info.experiment_id}.',
                    )

                log.debug(
                    f'MLFlowObjectStore using provided MLflow run {run_id=} '
                    f'for provided experiment {experiment_id=}',
                )
            else:
                # If no `run_id` is provided, create a new run in the provided experiment.
                run = self._mlflow_client.create_run(experiment_id)
                run_id = run.info.run_id

                log.debug(
                    f'MLFlowObjectStore using new MLflow run {run_id=} '
                    f'for provided experiment {experiment_id=}',
                )

        if experiment_id is None or run_id is None:
            raise ValueError('MLFlowObjectStore failed to initialize experiment and run ID.')

        return experiment_id, run_id

    @staticmethod
    def parse_dbfs_path(path: str) -> tuple[str, str, str]:
        """Parses a DBFS path to extract the MLflow experiment ID, run ID, and relative artifact path.

        The path is expected to be of the format
        `databricks/mlflow-tracking/<experiment_id>/<run_id>/artifacts/<artifact_path>`.

        Args:
            path (str): The DBFS path to parse.

        Returns:
            (str, str, str): (experiment_id, run_id, artifact_path)

        Raises:
            ValueError: If the path is not of the expected format.
        """
        if not path.startswith(MLFLOW_DBFS_PATH_PREFIX):
            raise ValueError(f'DBFS MLflow path should start with {MLFLOW_DBFS_PATH_PREFIX}. Got: {path}')

        # Strip `databricks/mlflow-tracking/` and split into
        # `<experiment_id>`, `<run_id>`, `'artifacts'`, `<relative_path>``
        subpath = path[len(MLFLOW_DBFS_PATH_PREFIX):]
        mlflow_parts = subpath.split('/', maxsplit=3)

        if len(mlflow_parts) != 4 or mlflow_parts[2] != 'artifacts':
            raise ValueError(
                f'Databricks MLflow artifact path expected to be of the format '
                f'{MLFLOW_DBFS_PATH_PREFIX}/<experiment_id>/<run_id>/artifacts/<relative_path>. '
                f'Found {path=}',
            )

        return mlflow_parts[0], mlflow_parts[1], mlflow_parts[3]

    def get_artifact_path(self, object_name: str) -> str:
        """Converts an object name into an MLflow relative artifact path.

        Args:
            object_name (str): The object name to convert. If the object name is a DBFS path beginning with
                ``MLFLOW_DBFS_PATH_PREFIX``, the path will be parsed to extract the MLflow relative artifact path.
                Otherwise, the object name is assumed to be a relative artifact path and will be returned as-is.
        """
        if object_name.startswith(MLFLOW_DBFS_PATH_PREFIX):
            experiment_id, run_id, object_name = self.parse_dbfs_path(object_name)
            if (experiment_id != self.experiment_id and experiment_id != MLFLOW_EXPERIMENT_ID_PLACEHOLDER):
                raise ValueError(
                    f'Object {object_name} belongs to experiment ID {experiment_id}, '
                    f'but MLFlowObjectStore is associated with experiment ID {self.experiment_id}.',
                )
            if (run_id != self.run_id and run_id != MLFLOW_RUN_ID_PLACEHOLDER):
                raise ValueError(
                    f'Object {object_name} belongs to run ID {run_id}, '
                    f'but MLFlowObjectStore is associated with run ID {self.run_id}.',
                )
        return object_name

    def get_dbfs_path(self, object_name: str) -> str:
        """Converts an object name to a full DBFS path."""
        artifact_path = self.get_artifact_path(object_name)
        return f'{MLFLOW_DBFS_PATH_PREFIX}{self.experiment_id}/{self.run_id}/artifacts/{artifact_path}'

    def get_uri(self, object_name: str) -> str:
        return 'dbfs:/' + self.get_dbfs_path(object_name)

    def upload_object(
        self,
        object_name: str,
        filename: Union[str, pathlib.Path],
        callback: Optional[Callable[[int, int], None]] = None,
    ):
        del callback  # unused
        from mlflow.exceptions import MlflowException

        # Extract relative path from DBFS path.
        artifact_path = self.get_artifact_path(object_name)
        artifact_base_name = os.path.basename(artifact_path)
        artifact_dir = os.path.dirname(artifact_path)

        # Since MLflow doesn't support uploading artifacts with a different base name than the local file,
        # create a temporary symlink to the local file with the same base name as the desired artifact name.
        filename = os.path.abspath(filename)

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_symlink_path = os.path.join(tmp_dir, artifact_base_name)
            os.symlink(filename, tmp_symlink_path)

            try:
                self._mlflow_client.log_artifact(self.run_id, tmp_symlink_path, artifact_dir)
            except MlflowException as e:
                _wrap_mlflow_exceptions(self.get_uri(object_name), e)

    def get_object_size(self, object_name: str) -> int:
        from mlflow.exceptions import MlflowException

        artifact = None
        try:
            artifact = self._get_artifact_info(object_name)
        except MlflowException as e:
            _wrap_mlflow_exceptions(self.get_uri(object_name), e)

        if artifact is not None:
            return artifact.file_size
        else:
            raise FileNotFoundError(f'Object {object_name} not found')

    def download_object(
        self,
        object_name: str,
        filename: Union[str, pathlib.Path],
        overwrite: bool = False,
        callback: Optional[Callable[[int, int], None]] = None,
    ) -> None:
        del callback  # unused
        from mlflow.exceptions import MlflowException

        # Since MlflowClient.download_artifacts only raises MlflowException with 500 Internal Error,
        # check for existence to surface a FileNotFoundError if necessary.
        artifact_path = self.get_artifact_path(object_name)
        artifact_info = self._get_artifact_info(object_name)
        if artifact_info is None:
            raise FileNotFoundError(f'Object {self.get_dbfs_path(artifact_path)} not found')

        filename = os.path.abspath(filename)
        if os.path.exists(filename) and not overwrite:
            raise FileExistsError(f'The file at {filename} already exists and overwrite is set to False.')

        # MLFLow doesn't support downloading artifacts directly to a specified filename, so instead
        # download to a temporary directory and then move the file to the desired location.
        with tempfile.TemporaryDirectory() as tmp_dir:
            try:
                self._mlflow_client.download_artifacts(
                    run_id=self.run_id,
                    path=artifact_path,
                    dst_path=tmp_dir,
                )
                tmp_path = os.path.join(tmp_dir, artifact_path)

                os.makedirs(os.path.dirname(filename), exist_ok=True)
                if overwrite:
                    os.replace(tmp_path, filename)
                else:
                    os.rename(tmp_path, filename)
            except MlflowException as e:
                _wrap_mlflow_exceptions(self.get_uri(artifact_path), e)

    def list_objects(self, prefix: Optional[str] = None) -> list[str]:
        """See :meth:`~composer.utils.ObjectStore.list_objects`.

        MLFlowObjectStore does not support listing objects with a prefix, so the ``prefix`` argument is ignored.
        """
        del prefix  # not supported for MLFlowObjectStore

        objects = []
        self._list_objects_helper(None, objects)
        return objects

    def _list_objects_helper(self, prefix: Optional[str], objects: list[str]) -> None:
        """Helper to recursively populate the full list of objects for ``list_objects``.

        Args:
            prefix (str | None): An artifact path prefix for artifacts to find.
            objects (list[str]): The list of DBFS object paths to populate.
        """
        from mlflow.exceptions import MlflowException

        artifact = None
        try:
            for artifact in self._mlflow_client.list_artifacts(self.run_id, prefix):
                if artifact.is_dir:
                    self._list_objects_helper(artifact.path, objects)
                else:
                    objects.append(artifact.path)
        except MlflowException as e:
            uri = '' if artifact is None else self.get_uri(artifact.path)
            _wrap_mlflow_exceptions(uri, e)

    def _get_artifact_info(self, object_name):
        """Get the :class:`~mlflow.entities.FileInfo` for the given object name.

        Args:
            object_name (str): The name of the object, either as an absolute DBFS path or a relative MLflow artifact path.

        Returns:
            Optional[FileInfo]: The :class:`~mlflow.entities.FileInfo` for the object, or None if it does not exist.
        """
        # MLflow doesn't support info for a singleton artifact, so we need to list all artifacts in the
        # parent path and find the one with the matching name.
        artifact_path = self.get_artifact_path(object_name)
        artifact_dir = os.path.dirname(artifact_path)
        artifacts = self._mlflow_client.list_artifacts(self.run_id, artifact_dir)
        for artifact in artifacts:
            if not artifact.is_dir and artifact.path == artifact_path:
                return artifact

        return None



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import datetime

import pytest

from composer.core.time import Time, Timestamp, TimeUnit


@pytest.mark.parametrize(
    'time_string,expected_value,expected_unit',
    [
        ['2iter', 2, TimeUnit.ITERATION],
        ['1ep', 1, TimeUnit.EPOCH],
        ['2ba', 2, TimeUnit.BATCH],
        ['3e10sp', 3 * 10**10, TimeUnit.SAMPLE],
        ['3_0e10sp', 30 * 10**10, TimeUnit.SAMPLE],
        ['4tok', 4, TimeUnit.TOKEN],
        ['4_000tok', 4000, TimeUnit.TOKEN],
        ['4_00_0tok', 4000, TimeUnit.TOKEN],
        ['0.5dur', 0.5, TimeUnit.DURATION],
        ['1h20m40s', 4840, TimeUnit.SECOND],
    ],
)
def test_time_parse(time_string: str, expected_value: int, expected_unit: TimeUnit):
    time = Time.from_timestring(time_string)
    assert time.value == expected_value
    assert time.unit == expected_unit


@pytest.mark.parametrize(
    'expected_timestring,time',
    [
        ['2iter', Time(2, TimeUnit.ITERATION)],
        ['1ep', Time(1, TimeUnit.EPOCH)],
        ['2ba', Time(2, TimeUnit.BATCH)],
        ['3sp', Time(3, TimeUnit.SAMPLE)],
        ['4tok', Time(4, TimeUnit.TOKEN)],
        ['0.5dur', Time(0.5, TimeUnit.DURATION)],
        ['6sec', Time(6, TimeUnit.SECOND)],
    ],
)
def test_to_timestring(expected_timestring: str, time: Time):
    assert time.to_timestring() == expected_timestring


def test_time_math():
    t1 = Time.from_timestring('1ep')
    t2 = Time.from_timestring('2ep')
    t3 = Time.from_timestring('3ep')
    t4 = Time.from_timestring('0.5dur')
    assert t1 + t2 == t3
    assert t2 - t1 == t1
    assert t1 - t2 == -t1
    assert t1 < t2
    assert t1 <= t2
    assert t2 > t1
    assert t2 >= t1
    assert t3 >= t3
    assert t3 <= t3
    assert t4 * t2 == t1
    assert 0.5 * t2 == t1
    assert t4 * 2 == Time.from_timestring('1dur')
    assert t1 / t2 == t4
    assert t2 / 2 == t1
    assert t3 % t3 == Time.from_timestring('0ep')
    assert t3 % t2 == t1


def test_invalid_math():
    t1 = Time.from_timestring('1ep')
    t2 = Time.from_timestring('1ba')

    with pytest.raises(RuntimeError):
        _ = t1 > t2

    with pytest.raises(RuntimeError):
        _ = t1 < t2

    with pytest.raises(RuntimeError):
        _ = t1 >= t2

    with pytest.raises(RuntimeError):
        _ = t1 <= t2

    with pytest.raises(RuntimeError):
        _ = t1 == t2

    with pytest.raises(RuntimeError):
        _ = t1 != t2

    with pytest.raises(RuntimeError):
        _ = t1 + t2

    with pytest.raises(RuntimeError):
        _ = t1 - t2

    with pytest.raises(RuntimeError):
        _ = t1 / t2

    with pytest.raises(RuntimeError):
        _ = t1 % t2

    with pytest.raises(RuntimeError):
        _ = t1 * t2


def test_time_from_input():
    expected = Time(1, TimeUnit.EPOCH)

    assert Time.from_input(expected) == expected
    assert Time.from_input('1ep') == expected
    assert Time.from_input(1, TimeUnit.EPOCH) == expected
    assert Time.from_input(1, 'ep') == expected

    with pytest.raises(RuntimeError):
        Time.from_input(None)  # type: ignore

    with pytest.raises(RuntimeError):
        Time.from_input([123])  # type: ignore

    with pytest.raises(RuntimeError):
        Time.from_input(1)


def test_time_repr():
    time = Time(1, 'tok')
    assert repr(time) == 'Time(1, TimeUnit.TOKEN)'
    assert eval(repr(time)) == time


def test_timestamp():
    timestamp = Timestamp()
    time = Time(10, 'ep')
    assert timestamp < time
    assert timestamp.get(time.unit) == Time.from_epoch(0)


def test_timestamp_update():
    timestamp = Timestamp(epoch=1)
    timestamp_2 = timestamp.copy(batch=2)
    assert timestamp_2.epoch == 1
    assert timestamp_2.batch == 2
    assert timestamp_2.sample == 0
    assert timestamp is not timestamp_2


def test_set_timestamp():
    timestamp = Timestamp(epoch_in_iteration=1)
    assert timestamp.epoch_in_iteration == 1
    timestamp.epoch_in_iteration = 2
    assert timestamp.epoch_in_iteration == 2


def test_timestamp_to_next_batch_epoch_iteration():
    timestamp = Timestamp()
    # Step batch 0 in epoch 0
    timestamp = timestamp.to_next_batch(10, 20, datetime.timedelta(seconds=5))
    assert timestamp.batch == 1
    assert timestamp.token_in_iteration == 20
    assert timestamp.batch_in_epoch == 1
    assert timestamp.sample == 10
    assert timestamp.sample_in_epoch == 10
    assert timestamp.token == 20
    assert timestamp.token_in_epoch == 20
    assert timestamp.total_wct == datetime.timedelta(seconds=5)
    assert timestamp.iteration_wct == datetime.timedelta(seconds=5)
    assert timestamp.epoch_wct == datetime.timedelta(seconds=5)
    assert timestamp.batch_wct == datetime.timedelta(seconds=5)

    # Finish epoch 0
    timestamp = timestamp.to_next_epoch(duration=datetime.timedelta(seconds=5))
    assert timestamp.epoch == 1
    assert timestamp.batch == 1
    assert timestamp.token_in_iteration == 20
    assert timestamp.batch_in_epoch == 0
    assert timestamp.sample == 10
    assert timestamp.sample_in_epoch == 0
    assert timestamp.token == 20
    assert timestamp.token_in_epoch == 0
    assert timestamp.total_wct == datetime.timedelta(seconds=10)
    assert timestamp.iteration_wct == datetime.timedelta(seconds=10)
    assert timestamp.epoch_wct == datetime.timedelta(seconds=0)
    assert timestamp.batch_wct == datetime.timedelta(seconds=0)

    # Step batch 0 in epoch 1
    timestamp = timestamp.to_next_batch(5, 0, datetime.timedelta(seconds=10))
    assert timestamp.epoch == 1
    assert timestamp.batch == 2
    assert timestamp.epoch_in_iteration == 1
    assert timestamp.token_in_iteration == 20
    assert timestamp.batch_in_epoch == 1
    assert timestamp.sample == 15
    assert timestamp.sample_in_epoch == 5
    assert timestamp.token == 20
    assert timestamp.token_in_epoch == 0
    assert timestamp.total_wct == datetime.timedelta(seconds=20)
    assert timestamp.iteration_wct == datetime.timedelta(seconds=20)
    assert timestamp.epoch_wct == datetime.timedelta(seconds=10)
    assert timestamp.batch_wct == datetime.timedelta(seconds=10)

    # Step batch 1 in epoch 1
    timestamp = timestamp.to_next_batch(5, 1, datetime.timedelta(seconds=10))
    assert timestamp.epoch == 1
    assert timestamp.batch == 3
    assert timestamp.token_in_iteration == 21
    assert timestamp.batch_in_epoch == 2
    assert timestamp.sample == 20
    assert timestamp.sample_in_epoch == 10
    assert timestamp.token == 21
    assert timestamp.token_in_epoch == 1
    assert timestamp.total_wct == datetime.timedelta(seconds=30)
    assert timestamp.iteration_wct == datetime.timedelta(seconds=30)
    assert timestamp.epoch_wct == datetime.timedelta(seconds=20)
    assert timestamp.batch_wct == datetime.timedelta(seconds=10)

    # Finish epoch 1
    timestamp = timestamp.to_next_epoch()
    assert timestamp.epoch == 2
    assert timestamp.batch == 3
    assert timestamp.epoch_in_iteration == 2
    assert timestamp.token_in_iteration == 21
    assert timestamp.batch_in_epoch == 0
    assert timestamp.sample == 20
    assert timestamp.sample_in_epoch == 0
    assert timestamp.token == 21
    assert timestamp.token_in_epoch == 0
    assert timestamp.total_wct == datetime.timedelta(seconds=30)
    assert timestamp.epoch_wct == datetime.timedelta(seconds=0)
    assert timestamp.batch_wct == datetime.timedelta(seconds=0)

    # Step batch 0 in epoch 2
    timestamp = timestamp.to_next_batch(5, 1, datetime.timedelta(seconds=10))
    assert timestamp.epoch == 2
    assert timestamp.batch == 4
    assert timestamp.epoch_in_iteration == 2
    assert timestamp.token_in_iteration == 22
    assert timestamp.batch_in_epoch == 1
    assert timestamp.sample == 25
    assert timestamp.sample_in_epoch == 5
    assert timestamp.token == 22
    assert timestamp.token_in_epoch == 1
    assert timestamp.total_wct == datetime.timedelta(seconds=40)
    assert timestamp.iteration_wct == datetime.timedelta(seconds=40)
    assert timestamp.epoch_wct == datetime.timedelta(seconds=10)
    assert timestamp.batch_wct == datetime.timedelta(seconds=10)

    # Finish iteration 0
    timestamp = timestamp.to_next_iteration()
    assert timestamp.iteration == 1
    assert timestamp.epoch == 2
    assert timestamp.batch == 4
    assert timestamp.epoch_in_iteration == 0
    assert timestamp.token_in_iteration == 0
    assert timestamp.batch_in_epoch == 0
    assert timestamp.sample == 25
    assert timestamp.sample_in_epoch == 0
    assert timestamp.token == 22
    assert timestamp.token_in_epoch == 0
    assert timestamp.total_wct == datetime.timedelta(seconds=40)
    assert timestamp.iteration_wct == datetime.timedelta(seconds=0)
    assert timestamp.epoch_wct == datetime.timedelta(seconds=0)
    assert timestamp.batch_wct == datetime.timedelta(seconds=0)


def test_timestamp_repr():
    timestamp = Timestamp()
    assert timestamp == eval(repr(timestamp))


@pytest.mark.parametrize('time_string', ['1.1iter', '1.5ep', '2.1ba', '3.2sp', '3.4tok', '0.1sec'])
def test_timestep_bad_strings(time_string: str):
    with pytest.raises(TypeError):
        Time.from_timestring(time_string)


@pytest.mark.parametrize('time_string', ['0.5dur', '1.0iter', '2.0ep', '3.000ba', '030.0sp', '30sec'])
def test_timestep_valid_strings(time_string: str):
    Time.from_timestring(time_string)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import dataclasses
from typing import Optional, Union

import numpy as np
import torch
import torch.nn.functional as F


@dataclasses.dataclass
class LowRankSolution:
    """Bundles tensors used by a factorized linear operator.

    The factorization always splits the operator into two smaller linear
    operators. The first takes in input of the original shape and embeds it
    in a lower-dimensional space. The second maps this lower-dimensional space
    to the original output space.

    Args:
        Wa (torch.Tensor, optional): First linear operation in the
            factorized approximation. For a
            factorized linear operation, ``Wa`` is a matrix. For a factorized
            convolution, ``Wa`` matches the shape of the convolution's
            original weight parameter, except along the channel axis.
        Wb (torch.Tensor, optional): Second linear operation in the
            factorized approximation. Shape
            is such that composing ``Wb`` with ``Wb`` yields an output of
            the same size as the original operation.
        bias (torch.Tensor, optional): Vector added to the output of
            the second linear operation.
        rank (int, optional): Output dimensionality (channels or features) of
            the first linear operation, and input dimensionality of the second
            input operation. Default: ``-1``.
        nmse (float, optional): Normalized mean squared error obtained during
            the optimization procedure used to derive ``Wa``, ``Wb``, and
            ``bias``. This is equal to the raw mean squared error between
            the factorized approximation's output and the original output,
            divided by the variance of the original output. A value of 0
            means no error was introduced, and a value of 1 corresponds to
            capturing the output no better than chance. Default: ``0.0``.
    """
    Wa: Optional[torch.Tensor] = None
    Wb: Optional[torch.Tensor] = None
    bias: Optional[torch.Tensor] = None
    rank: int = -1
    nmse: float = 0


def _lstsq(A: torch.Tensor, B: torch.Tensor) -> torch.Tensor:
    if A.shape[0] != B.shape[0]:
        raise RuntimeError(f'A has different number of rows than B! A.shape = {A.shape}, B.shape = {B.shape}')
    if A.ndim != 2:
        raise RuntimeError('A is not a rank 2 tensor: has shape', A.shape)
    if B.ndim != 2:
        raise RuntimeError('B is not a rank 2 tensor: has shape', A.shape)

    # TODO more intelligence regarding choice of lstsq `driver` arg
    return torch.linalg.lstsq(A, B).solution


def _nmse(Y: torch.Tensor, Y_hat: torch.Tensor) -> float:
    diffs = Y.detach() - Y_hat.detach()
    return float((diffs * diffs).mean() / Y.var())


def _svd_initialize(Wa: torch.Tensor, Wb: Optional[torch.Tensor], k: int) -> tuple[torch.Tensor, torch.Tensor]:
    if Wb is None:
        W = Wa
    else:
        W = Wa @ Wb

    # TODO rank k randomized svd if k small enough
    U, s, Vt = torch.linalg.svd(W, full_matrices=False)
    Wa = U[:, :k]
    Wb = Vt[:k]

    # scale matrices equally for numerical "load-balancing"
    s_sqrt = torch.sqrt(s[:k])  # s is already a vector, not mat
    Wa *= s_sqrt
    Wb *= s_sqrt.reshape(-1, 1)
    return Wa, Wb


def factorize_matrix(
    X: torch.Tensor,
    Y: torch.Tensor,
    Wa: torch.Tensor,
    Wb: Optional[torch.Tensor] = None,
    bias: Optional[torch.Tensor] = None,
    rank: Union[int, float] = .25,
    n_iters: int = 3,
) -> LowRankSolution:
    """Approximates a matrix by factorizing it into a product of two smaller matrices.

    Given a matrix ``W`` of shape ``[D, M]``, a bias vector of length ``M``,
    and a target rank ``rank < D``, returns a solution ``(Wa, Wb, new_bias)`` of
    tensors of shapes ``[N, rank]``, ``[rank, D]``, and ``M``, respectively.
    These tensors are chosen so as to minimize:

    :math:`||` ``Y - (X @ Wa @ Wb + new_bias)`` :math:`||_F`,

    where ``Y = X @ W + bias``, ``@`` denotes matrix multiplication,
    ``new_bias`` broadcasts along the row dimension,
    and :math:`||\\cdot||_F` denotes the sum of squared elements.
    In the case that rows of ``X`` correspond to samples
    from some distribution, this amounts to minimizing the expected mean
    squared error in the output.

    The input matrix can either be a single matrix ``W`` or a pair of matrices
    ``(Wa, Wb)``. The latter case corresponds to using a matrix ``W = Wa @ Wb``
    that has already been factorized and is supported in order to facilitate
    progressively decreasing the rank of the matrix.

    Args:
        X (torch.Tensor): Input used to evaluate the quality of the approximation.
            Shape is ``[N, D]``, where ``N`` is often the number of input samples and
            ``D`` is the dimensionality of each sample.
        Y (torch.Tensor): Output of applying the original matrix to ``X``.
            Must have shape ``[N, M]`` for some ``M``.
        Wa (torch.Tensor): Either the matrix to be factorized, or the first of the two smaller
            matrices in the already-factorized representation of this matrix.
            Must be of shape ``[D, M]`` in the former case and shape ``[D, d]``
            in the latter, for some ``d < D``.
        Wb (torch.Tensor, optional): If present, ``Wa`` is interpreted
            as the first of two smaller matrices, and ``Wb`` is taken to be the second.
            Must be of shape ``[d, M]``.
        bias (torch.Tensor, optional): A vector added to the output after
            performing the matrix product with X.
        rank (int | float, optional): the number of columns in the latent representation of X.
            Default: ``.25``.
        n_iters (int, optional): number of iterations used in the optimization process. Higher
            numbers yield lower mean squared error, though there are usually
            diminishing returns after a handful of iterations. Default: ``3``.

    Returns:
        LowRankSolution: A solution of rank ``rank`` that approximates the original convolution operation.
    """
    X = X.detach()
    Y = Y.detach()
    Wa = Wa.detach()
    Wb = Wb.detach() if Wb is not None else None
    if rank < 1:
        # fraction of input dimensionality (or current rank, if smaller)
        rank = min(int(rank * X.shape[1]), Wa.shape[1])
    k = int(rank)

    solution = LowRankSolution()

    original_bias = None
    if bias is not None:
        original_bias = bias.detach()
        Y = Y - original_bias
        solution.bias = original_bias

    # if requested latent rank is greater than or equal to either
    # input rank or output rank, factorization is counterproductive, so
    # return a single matrix
    if k >= X.shape[1] or k >= Y.shape[1]:
        Wa = _lstsq(X, Y)
        solution.Wa = Wa
        solution.rank = -1
        return solution

    # if requested latent rank is greater than current latent rank,
    # skip the factorization
    if k >= Wa.shape[1]:
        solution.Wa = Wa
        solution.Wb = Wb
        solution.rank = -1
        return solution

    Wa, Wb = _svd_initialize(Wa, Wb, k)

    Ya = _lstsq(X, Y)
    for _ in range(n_iters):
        # update Wb
        Xb = X @ Wa
        Yb = Y
        Wb = _lstsq(Xb, Yb)

        # update Wa
        # We need to solve (AXB = Y) <=> (AX = B.I @ Y) not (AX = BY).
        # Since X and Y are constants, we can precompute pinv(A) @ Y.
        # We then have:
        #   pinv(A) @ A @ X @ B = pinv(A) @ Y
        #   (A.T@A).I @ A.T @ A @ X @ B = pinv(A) @ Y
        #   X @ B = pinv(A) @ Y
        #   Y.T @ pinv(A).T = B.T @ X.T
        # then we just solve for X.T:
        #   B.T @ X.T = Y.T @ pinv(A).T
        # also, note that pinv(A) @ Y = lstsq(A, Y); this makes sense;
        # means that targets for XB are the optimal coeffs mapping A to Y
        # also, observe that AXB = Y is using X and Y as variable to solve
        # for and targets, not the X and Y vars we have in this function
        Xa = Wb
        Wa_T = _lstsq(Xa.T, Ya.T)
        Wa = Wa_T.T

    solution.Wa = Wa
    solution.Wb = Wb
    solution.rank = k
    Y_hat = (X @ Wa) @ Wb

    bias = (Y - Y_hat).mean(dim=0)
    if original_bias is not None:
        bias += original_bias
    solution.bias = bias

    Y_hat += bias
    solution.nmse = _nmse(Y, Y_hat)

    return solution


def _activations_conv2d_to_mat(
    activations,
    kernel_size,
    padding=0,
    padding_mode='zeros',
    stride=1,
    dilation=1,
    groups=1,
):
    if np.max(stride) > 1:
        raise NotImplementedError(f'Stride != 1 not implemented; got {stride}')
    if np.max(dilation) > 1:
        raise NotImplementedError(f'Dilation != 1 not implemented; got {dilation}')
    if groups != 1:
        raise NotImplementedError(f'Groups != 1 not implemented; got {groups}')
    if np.max(padding) > 0 and padding_mode.lower() != 'zeros':
        if not isinstance(padding, list):
            padding = [padding]

        activations = F.pad(activations, pad=padding, mode=padding_mode)
        padding = 0
    # always default to stride=1 to maximize amount of data we get here
    # TODO downsample in batch size dim or use stride > 1 if it looks like
    # materializing full matrix will OOM
    ret = F.unfold(activations, kernel_size=kernel_size, padding=padding)
    ret = ret.transpose(1, 2)  # batch_sz, n_positions, fan_in
    return ret.reshape(-1, ret.shape[2])  # batch_sz * n_positions, fan_in


def _weights_conv2d_to_mat(weights: torch.Tensor):
    return weights.reshape(weights.shape[0], -1).T  # fan_in, out_channels


def _mat_to_weights_conv2d(mat: Optional[torch.Tensor], kernel_size) -> Optional[torch.Tensor]:
    if mat is None:
        return None
    w = mat.T  # fan_in, out_channels -> out_channels, fan_in
    # XXX(nchw) This might silently do the wrong thing with nhwc layout
    return w.reshape(w.shape[0], -1, *kernel_size)


def factorize_conv2d(
    X: torch.Tensor,
    Wa: torch.Tensor,
    Wb: Optional[torch.Tensor] = None,
    rank: Union[int, float] = .25,
    biasA: Optional[torch.Tensor] = None,
    biasB: Optional[torch.Tensor] = None,
    n_iters=3,
    **conv2d_kwargs,
) -> LowRankSolution:
    """Approximates a :math:`K \\times K` convolution by factorizing it into a
    :math:`K \\times K` convolution  with fewer channels followed by a
    :math:`1 \\times 1` convolution.

    Given a convolutional weight tensor ``W`` for a 2d convolution of shape
    ``[out_channels, in_channels, k_h, k_w]`` and a vector ``bias`` of length
    ``out_channels``, returns a triple ``(Wa, Wb, new_bias)`` of
    tensors with shapes ``[rank, in_channels, k_h, k_w]``,
    ``[out_channels, rank, 1, 1]``, and ``[out_channels]``, respectively.
    ``Wa``, ``Wb``, and ``new_bias`` are chosen so as to minimize:

        :math:`||` ``(W * X + bias) - (Wb * (Wa * X) + new_bias)`` :math:`||_F`,

    where :math:`*` denotes convolution, ``bias`` broadcasts along all
    non-channel dimensions, and :math:`||\\cdot||_F` denotes the sum of
    squared elements.

    Similar to :func:`.factorize_matrix`, this function allows passing in an
    already-factorized weight tensor in order to enable progressive
    factorization. In this case, the single tensor ``W`` is replaced with
    a similar ``(Wa, Wb)`` pair as the output, though not necessarily with
    the same rank.

    Args:
        X (torch.Tensor): A tensor of shape ``[N, in_channels, H, W]``, for some
            ``N``, ``H``, and ``W``.
        Wa (torch.Tensor): The first weight tensor to convolve with ``X``. If
            ``Wb`` is not provided, must be of shape
            ``[out_channels, in_channels, k_h, k_w]``. Otherwise, must be of
            shape ``[original_rank, in_channels, k_h, k_w]`` for some
            ``original_rank < min(in_channels, out_channels)``.
        Wb (torch.Tensor, optional): The second weight tensor to convolve.
            with the input. If provided, must be of shape ``[out_channels, original_rank, 1, 1]``.
        rank (int | float, optional): number of channels in the latent representation of ``X``.
            Default: ``.25``.
        biasA (torch.Tensor, optional): Optional vector of biases. If ``Wb`` is
            ``None``, must have length ``out_channels``. Otherwise must have length
            ``original_rank``.
        biasB (torch.Tensor, optional): If provided, must have length ``out_channels``.
        n_iters (int, optional): number of iterations used in the optimization process.
            Higher numbers yield lower mean squared error, though there are usually
            diminishing returns after a handful of iterations. Default: ``3``.
        **conv2d_kwargs: Arguments such as ``padding``, ``stride``,
            ``dilation``, ``groups``, etc used in the original convolution. If
            these are not provided, the factorized tensors might not preserve
            the function computed by the original weight tensor as well.
            Note that not all combinations of arguments are supported.

    Returns:
        LowRankSolution: A solution of rank ``rank`` that approximates the original convolution operation.

    Raises:
        RuntimeError:
            If ``biasB`` is provided but not ``Wb`` is not.
        NotImplementedError:
            if ``conv2d_kwargs['dilation'] != 1`` or ``conv2d_kwargs['groups'] != 1``.
    """
    X = X.detach()
    Wa = Wa.detach()

    kernel_size = Wa.shape[2:]
    X_mat = _activations_conv2d_to_mat(X, kernel_size=kernel_size, **conv2d_kwargs)
    Wa = _weights_conv2d_to_mat(Wa)
    # NOTE: we compute outputs ourselves, instead of having an arg for them,
    # since 1) we ignore input stride, and 2) any other intermediate ops
    # or other discrepancies between user's actual settings and args they pass
    # would either cause errors or silently mess up the regression
    Y_mat = (X_mat @ Wa)
    if biasA is not None:
        biasA = biasA.detach()
        Y_mat += biasA

    if Wb is not None:
        Wb = Wb.detach()
        Wb = _weights_conv2d_to_mat(Wb)
        Y_mat = Y_mat @ Wb

        if biasB is not None:
            biasB = biasB.detach()
            Y_mat += biasB
    elif biasB is not None:
        # fail fast if user passes in inconsistent combination of args
        raise RuntimeError('Got biasB, but Wb=None; cannot apply bias')

    ret = factorize_matrix(X_mat, Y_mat, Wa, Wb, rank=rank, n_iters=n_iters)

    # now we need to convert from two matrices to one kxk conv kernel and one
    # 1x1 conv kernel. Here's why the 2nd has to be a 1x1: if it were instead
    # k'xk' for some k' > 1, we would either be doing k'^2 as much work
    # for fixed embedding size at each pixel, or we'd be need to have the
    # intermediate embeddings be 1/k'^2 as large. In the latter case, we'd
    # lose a lot of representational capacity. Also, the first op has to match
    # the kernel size of the original conv or the shapes don't work out.
    assert ret.Wa is not None
    ret.Wa = _mat_to_weights_conv2d(ret.Wa, kernel_size=kernel_size)
    ret.Wb = _mat_to_weights_conv2d(ret.Wb, kernel_size=(1, 1))

    return ret



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""ChannelsLast algorithm."""

from __future__ import annotations

import logging
from typing import Optional

import torch

from composer.core import Algorithm, Event, State
from composer.loggers import Logger

log = logging.getLogger(__name__)

__all__ = ['ChannelsLast', 'apply_channels_last']


def apply_channels_last(model: torch.nn.Module) -> None:
    """Changes the memory format of the model to `torch.channels_last <https://\\
    pytorch.org/tutorials/intermediate/memory_format_tutorial.html>`_.

    This usually yields improved GPU utilization.

    Args:
        model (torch.nn.Module): The model or module to modify.
    """
    model.to(memory_format=torch.channels_last)  # type: ignore


class ChannelsLast(Algorithm):
    """Changes the memory format of the model to `torch.channels_last <https://\\
    pytorch.org/tutorials/intermediate/memory_format_tutorial.html>`_. This usually improves GPU utilization.

    Runs on :attr:`.Event.INIT``, so it can set the memory format before the model is DDP wrapped.
    Has no hyperparameters.

    Example:
        .. testcode::

            from composer.algorithms import ChannelsLast
            algorithm = ChannelsLast()
            trainer = Trainer(
                model=model,
                train_dataloader=train_dataloader,
                eval_dataloader=eval_dataloader,
                max_duration="1ep",
                algorithms=[algorithm],
                optimizers=[optimizer]
            )
    """

    def __init__(self) -> None:
        # ChannelsLast takes no arguments
        pass

    def match(self, event: Event, state: State) -> bool:
        del state  # unused
        return event == Event.INIT

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        del event, logger  # unused
        # TODO: Double check model is moved to cuda with device type
        apply_channels_last(state.model)

        log.info(f'Model {state.model.__class__.__name__} changed to channels_last format.')



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Replaces model with a dummy model of type :class:`NoOpModelClass`.

The algorithm runs on :attr:`Event.INIT`. It replaces the model in the state with
a :class:`.NoOpModelClass` and then updates the parameters in the optimizer
through module surgery.

A dummy model can helpful for profiling the dataloader by eliminating the work
necessary to compute model outputs.
"""

from composer.algorithms.no_op_model.no_op_model import NoOpModel as NoOpModel
from composer.algorithms.no_op_model.no_op_model import NoOpModelClass as NoOpModelClass

__all__ = ['NoOpModel', 'NoOpModelClass']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

# Written by Gihyun Park, Junyeol Lee, and Jiwon Seo

import logging
import warnings
from typing import Optional, Type

import numpy as np
import torch

from composer.algorithms.warnings import NoEffectWarning
from composer.core import Algorithm, Event, State
from composer.loggers import Logger
from composer.utils import module_surgery

log = logging.getLogger(__name__)


class GyroDropoutLayer(torch.nn.Module):

    def __init__(self, iters_per_epoch: int, max_epoch: int, p: float, sigma: int, tau: int):
        super(GyroDropoutLayer, self).__init__()

        self.iters_per_epoch = iters_per_epoch
        self.max_epoch = max_epoch
        self.p = p
        self.sigma = sigma
        self.tau = tau
        self.preselect_masks = torch.empty(0, 0)
        self.dropout_mask = torch.empty(0, 0)
        self.selected_masks = torch.empty(0, 0)
        self.training_step = 0
        self.iter_num = 0

    def forward(self, x):
        if self.training:
            if self.training_step == 0:
                is_cuda_tensor = x.is_cuda

                if is_cuda_tensor:
                    self.preselect_masks = (torch.rand(self.sigma, x.shape[1]) > self.p).float().to('cuda')
                else:
                    self.preselect_masks = (torch.rand(self.sigma, x.shape[1]) > self.p).float()

                # Below simplified from: (iters_per_epoch*max_epoch*batch_size/sigma) / (batch_size/self.tau)
                self.iter_num = int(self.iters_per_epoch * self.max_epoch / self.sigma) * self.tau

            if self.training_step % self.iter_num == 0:
                pick_idx = np.random.choice(self.sigma, self.tau)
                self.selected_masks = self.preselect_masks[pick_idx]

            self.dropout_mask = torch.repeat_interleave(self.selected_masks, x.shape[0] // self.tau, dim=0)

            self.training_step += 1

            return x * self.dropout_mask * (1 / (1 - self.p))
        else:
            return x


def from_Dropout(
    iters_per_epoch: int,
    epoch: int,
    p: float,
    sigma: int,
    tau: int,
    layer: torch.nn.Module,
    module_index: int,
):
    """Defines a replacement policy from a `torch.nn.Dropout` to a 'GyroDropout`"""

    return GyroDropoutLayer(iters_per_epoch, epoch, p, sigma, tau)


def apply_gyro_dropout(
    model: torch.nn.Module,
    iters_per_epoch: int,
    max_epoch: int,
    p: float,
    sigma: int,
    tau: int,
) -> None:
    """Replaces all instances of `torch.nn.Dropout` with a `GyroDropout`.

    By masking Dropout layer, this usually improves accuracy.
    """

    # prepare the replacement policy and perform replacement
    from functools import partial
    policy: dict[Type[torch.nn.Module], module_surgery.ReplacementFunction] = {
        torch.nn.Dropout: partial(from_Dropout, iters_per_epoch, max_epoch, p, sigma, tau),
    }
    replaced_instances = module_surgery.replace_module_classes(module=model, policies=policy)
    if len(replaced_instances) == 0:
        warnings.warn(
            NoEffectWarning(
                'No instances of `torch.nn.Dropout` were found, and therefore, there were no modules to replace.',
            ),
        )
    log.info(f'Successfully replaced {len(replaced_instances)} of dropout with a Gyro dropout.')


class GyroDropout(Algorithm):
    """Replaces all instances of `torch.nn.Dropout` with a `GyroDropout`.

    By masking Dropout layer, this usually improves accuracy.

    Args:
        p (float, optional): Float number of ratio to dropout.
            Default: ``0.5``.
        sigma (int, optional): the number of total pre-selected subnetwork
            Default: ``256``.
        tau (int, optional): the number of concurrently scheduled subnetworks in an iteration
            Default: ``16``.

    Example:

        .. testcode::

           from composer.algorithms import GyroDropout

           algorithm = GyroDropout(p=0.5, sigma=256, tau=16)
           trainer = Trainer(
               model=model,
               train_dataloader=train_dataloader,
               max_duration="100ep",
               algorithms=[algorithm],
               optimizers=[optimizer]
           )
    """

    def __init__(self, p: float = 0.5, sigma: int = 256, tau: int = 16):
        self.p = p
        self.sigma = sigma
        self.tau = tau

        warnings.warn(
            'GyroDropout is not implemented in a way that allows correct resumption from checkpoint, which may lead to incorrect behavior.',
        )

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'

    @staticmethod
    def required_on_load() -> bool:
        return True

    def match(self, event: Event, state: State) -> bool:
        del state
        return event == Event.FIT_START

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        del event, logger

        assert state.dataloader_len is not None
        assert state.max_duration is not None

        apply_gyro_dropout(
            model=state.model,
            iters_per_epoch=state.dataloader_len.value,
            max_epoch=state.max_duration.value,
            p=self.p,
            sigma=self.sigma,
            tau=self.tau,
        )



# Copyright 2024 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Module for checkpointing API."""

from composer.checkpoint.state_dict import (
    get_metadata_state_dict,
    get_model_state_dict,
    get_optim_state_dict,
    get_resumption_state_dict,
)

__all__ = [
    'get_model_state_dict',
    'get_optim_state_dict',
    'get_metadata_state_dict',
    'get_resumption_state_dict',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""The models module contains the :class:`.ComposerModel` base class along with reference
implementations of many common models. Additionally, it includes task-specific convenience
:class:`.ComposerModel`\\s that wrap existing Pytorch models with standard forward passes
and logging to enable quick interaction with the :class:`.Trainer`.

See :doc:`Composer Model </composer_model>` for more details.
"""

from composer.models.base import ComposerModel
from composer.models.huggingface import HuggingFaceModel, write_huggingface_pretrained_from_composer_checkpoint
from composer.models.initializers import Initializer
from composer.models.tasks import ComposerClassifier

__all__ = [
    'ComposerModel',
    'HuggingFaceModel',
    'write_huggingface_pretrained_from_composer_checkpoint',
    'Initializer',
    'ComposerClassifier',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Monitor train and eval images."""
from typing import Any, Callable, Sequence, Union

import torch

from composer.core import Callback, State, Time, TimeUnit
from composer.loggers import Logger
from composer.loss.utils import infer_target_type

__all__ = ['ImageVisualizer']


class ImageVisualizer(Callback):
    """Logs image inputs and optionally outputs.

    This callback triggers at a user defined interval, and logs a sample of input (optionally also segmentation masks)
    images under the ``Images/Train`` and ``Image/Eval`` keys.

    Example:
        .. doctest::

            >>> from composer import Trainer
            >>> from composer.callbacks import ImageVisualizer
            >>> # constructing trainer object with this callback
            >>> trainer = Trainer(
            ...     model=model,
            ...     train_dataloader=train_dataloader,
            ...     eval_dataloader=eval_dataloader,
            ...     optimizers=optimizer,
            ...     max_duration="1ep",
            ...     callbacks=[ImageVisualizer()],
            ... )

    The images are logged by the :class:`.Logger` to the following key(s) as described
    below.

    +---------------------------------------------+---------------------------------------+
    | Key                                         | Logged data                           |
    +=============================================+=======================================+
    | ``Images/Train``                            |   Sampled examples of train images    |
    +---------------------------------------------+---------------------------------------+
    | ``Images/Eval``                             |   Sampled examples of eval images     |
    +---------------------------------------------+---------------------------------------+

        .. note::
            This callback only works with wandb and Neptune logging for now.

    Args:
        interval (int | str | Time, optional): Time string specifying how often to log train images. For example, ``interval='1ep'``
            means images are logged once every epoch, while ``interval='100ba'`` means images are logged once every 100
            batches. Eval images are logged once at the start of each eval. Default: ``"100ba"``.
        mode (str, optional): How to log the image labels. Valid values are ``"input"`` (input only)
            and "segmentation" which also saves segmentation masks. Default: ``"input"``.
        num_images (int, optional): Number of images to log. Should be less than or equal to than the microbatch size.
            If there are not enough images in the microbatch, all the images in the microbatch will be logged.
            Default: ``8``.
        channels_last (bool, optional): Whether the image channel dimension is the last dimension. Default: ``False``.
        input_key (str | int | tuple[Callable, Callable] | Any, optional): A key that indexes to the input
            from the batch. Can also be a pair of get and set functions, where the getter
            is assumed to be first in the pair.  The default is 0, which corresponds to any sequence, where the first
            element is the input. Default: ``0``.
        target_key (str | int | tuple[Callable, Callable] | Any, optional): A key that indexes to the target
            from the batch. Can also be a pair of get and set functions, where the getter
            is assumed to be first in the pair. The default is 1, which corresponds to any sequence, where the second
            element is the target. Default: ``1``.
    """

    def __init__(
        self,
        interval: Union[int, str, Time] = '100ba',
        mode: str = 'input',
        num_images: int = 8,
        channels_last: bool = False,
        input_key: Union[str, int, tuple[Callable, Callable], Any] = 0,
        target_key: Union[str, int, tuple[Callable, Callable], Any] = 1,
    ):
        self.mode = mode
        self.num_images = num_images
        self.channels_last = channels_last
        self.input_key = input_key
        self.target_key = target_key

        # Check that the output mode is valid
        if self.mode.lower() not in ['input', 'segmentation']:
            raise ValueError(f'Invalid mode: {mode}')

        # Check that the interval timestring is parsable and convert into time object
        self.interval = Time.from_input(interval, TimeUnit.BATCH)

        # Verify that the interval has supported units
        if self.interval.unit not in [TimeUnit.BATCH, TimeUnit.EPOCH]:
            raise ValueError(
                f'Invalid time unit for parameter interval: '
                f'{self.interval.unit}',
            )

        self.last_train_time_value_logged = -1

    def _log_inputs(self, state: State, logger: Logger, data_name: str):
        inputs = state.batch_get_item(key=self.input_key)
        # Verify inputs is a valid shape for conversion to an image
        if _check_for_image_format(inputs):
            inputs = _make_input_images(inputs, self.num_images)
            logger.log_images(inputs, name=data_name, use_table=True, channels_last=self.channels_last)

    def _log_segmented_inputs(self, state: State, logger: Logger, data_name: str):
        inputs = state.batch_get_item(key=self.input_key)
        targets = state.batch_get_item(key=self.target_key)
        outputs = state.outputs
        assert isinstance(outputs, torch.Tensor)

        images, masks = _make_segmentation_images(inputs, targets, outputs, self.num_images, self.channels_last)
        logger.log_images(images, masks=masks, name=data_name, channels_last=self.channels_last, use_table=True)

    def before_forward(self, state: State, logger: Logger):
        current_time_value = state.timestamp.get(self.interval.unit).value
        if self.mode.lower(
        ) == 'input' and current_time_value % self.interval.value == 0 and current_time_value != self.last_train_time_value_logged:
            self.last_train_time_value_logged = current_time_value
            self._log_inputs(state, logger, 'Images/Train')

    def eval_before_forward(self, state: State, logger: Logger):
        if self.mode.lower() == 'input' and state.eval_timestamp.get(TimeUnit.BATCH).value == 0:
            self._log_inputs(state, logger, 'Images/Eval')

    def before_loss(self, state: State, logger: Logger):
        current_time_value = state.timestamp.get(self.interval.unit).value
        if self.mode.lower(
        ) == 'segmentation' and current_time_value % self.interval.value == 0 and current_time_value != self.last_train_time_value_logged:
            self.last_train_time_value_logged = current_time_value
            self._log_segmented_inputs(state, logger, 'Images/Train')

    def eval_after_forward(self, state: State, logger: Logger):
        if self.mode.lower() == 'segmentation' and state.eval_timestamp.get(TimeUnit.BATCH).value == 0:
            self._log_segmented_inputs(state, logger, 'Images/Eval')


def _make_input_images(inputs: torch.Tensor, num_images: int):
    if inputs.shape[0] < num_images:
        num_images = inputs.shape[0]
    images = inputs[0:num_images].data.cpu().numpy()
    return images


def _make_segmentation_images(
    inputs: torch.Tensor,
    targets: torch.Tensor,
    outputs: Union[torch.Tensor, Sequence[torch.Tensor]],
    num_images: int,
    channels_last: bool = False,
):
    if isinstance(outputs, Sequence):
        outputs = torch.stack(list(outputs))
    if min([inputs.shape[0], targets.shape[0], outputs.shape[0]]) < num_images:
        num_images = min([inputs.shape[0], targets.shape[0], outputs.shape[0]])

    images = inputs[0:num_images].data.cpu().numpy()

    targets = targets[0:num_images]
    outputs = outputs[0:num_images]
    # Ensure the targets are in the expected format
    if infer_target_type(outputs, targets) == 'one_hot':
        if channels_last:
            targets = targets.argmax(dim=-1).data.cpu().numpy()
        else:
            targets = targets.argmax(dim=1).data.cpu().numpy()
    else:
        targets = targets.data.cpu().numpy()
    # Convert the outputs to the expected format
    if channels_last:
        num_classes = outputs.shape[-1]
        outputs = outputs.argmax(dim=-1).cpu().numpy()
    else:
        num_classes = outputs.shape[1]
        outputs = outputs.argmax(dim=1).cpu().numpy()
    # Adjust targets such that negative values are mapped to one higher than the maximum class
    targets[targets < 0] = num_classes

    return images, {'prediction': outputs, 'ground_truth': targets}


def _check_for_image_format(data: torch.Tensor) -> bool:
    return data.ndim in [3, 4] and data.numel() > data.shape[0]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Replaces batch normalization modules with `Ghost Batch Normalization <https://arxiv.org/abs/1705.08741>`_ modules
that simulate the effect of using a smaller batch size.

See :class:`~composer.algorithms.GhostBatchNorm` or the :doc:`Method Card </method_cards/ghost_batchnorm>` for details.
"""

from composer.algorithms.ghost_batchnorm.ghost_batchnorm import GhostBatchNorm as GhostBatchNorm
from composer.algorithms.ghost_batchnorm.ghost_batchnorm import apply_ghost_batchnorm as apply_ghost_batchnorm

__all__ = ['GhostBatchNorm', 'apply_ghost_batchnorm']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Distributed training."""

from composer.distributed.deepspeed import fix_batch_precision_for_deepspeed, parse_deepspeed_config
from composer.distributed.dist_strategy import (
    DDPSyncStrategy,
    ddp_sync_context,
    prepare_ddp_module,
    prepare_fsdp_module,
    prepare_tp_module,
)

__all__ = [
    'fix_batch_precision_for_deepspeed',
    'parse_deepspeed_config',
    'DDPSyncStrategy',
    'ddp_sync_context',
    'prepare_ddp_module',
    'prepare_fsdp_module',
    'prepare_tp_module',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Eval client for local evaluation."""
import logging
import multiprocessing
import os
import subprocess
import textwrap
import types

from composer.utils import dist
from composer.utils.eval_client.eval_client import EvalClient

__all__ = ['LocalEvalClient']
log = logging.getLogger(__name__)

TIMEOUT = 5  # in seconds


class LocalEvalClient(EvalClient):
    """Utility for creating a client for and invoking local evaluations."""

    def invoke(self, payload: list[list[list[dict[str, str]]]]) -> list[list[list[bool]]]:
        """Invoke a batch of provided payloads for code evaluations."""
        ret = []
        for prompt_group in payload:
            ret_prompt_group = []
            for generation_group in prompt_group:
                ret_generation_group = []
                for test_case in generation_group:
                    ret_generation_group.append(self.invoke_helper(test_case))
                ret_prompt_group.append(ret_generation_group)
            ret.append(ret_prompt_group)
        return ret

    def invoke_helper(self, payload: dict[str, str]) -> bool:
        """Invoke a provided dictionary payload to a multiprocessing subprocess that performs code eval."""
        ret = multiprocessing.Value('b', 0)  # Store result of test case in shared memory
        p = multiprocessing.Process(
            target=self.update_offline_helper,
            args=(
                payload['code'],
                payload['input'],
                payload['output'],
                payload['entry_point'],
                payload['language'],
                ret,
            ),
        )  # Evaluate test case in an independent subprocess
        p.start()
        p.join(TIMEOUT)  # wait for timeout to terminate
        p.terminate()
        return bool(ret.value)  # pyright: ignore[reportGeneralTypeIssues]

    def update_offline_helper(
        self,
        code_gen: str,
        test_input: str,
        test_output: str,
        entry_point: str,
        language: str,
        val: multiprocessing.Value,  # type: ignore
    ):
        """Helper function to evaluate test case in a subprocess.

        This function compiles the code generation,
        and runs the function from the entry point, before running the test input through the function and
        checking it against the test output.

        Args:
            code_gen (str): The code generation to be evaluated.
            test_input (str): The input of the test case
            test_output (str): The output of the test case
            entry_point (str): The name of the function to call
            language (str): The language of the code generation
            val (multiprocessing.Value): The value in which to save the final value of the test case
        """
        rank = dist.get_global_rank()
        if language == 'python':
            mod = types.ModuleType('test_module')
            val.value = 0
            result = None
            expected_result = None
            try:
                exec(code_gen, mod.__dict__)

                result = mod.__dict__[entry_point](*eval(test_input))
                syntax_compiled = True
                try:
                    expected_result = eval(test_output)
                except:
                    expected_result = test_output

            except Exception as _:
                syntax_compiled = False  # don't print out warning b/c this can happen frequently

            if syntax_compiled:
                val.value = int(result == expected_result)
            else:
                val.value = 0
        elif language == 'c++':
            # since we can't run tostring based comparisons in C++, initialize a series of overloaded equality functions
            prefix = '''\
                #include <iostream>
                #include <vector>
                #include <string>
                #include <map>
                #include <math.h>
                using namespace std;

                bool issame(vector<string> a,vector<string>b){
                    if (a.size()!=b.size()) return false;
                    for (int i=0;i<a.size();i++)
                    {
                        if (a[i]!=b[i]) return false;
                    }
                    return true;
                }

                bool issame(float a, float b){
                    return abs(a - b) < 1e-4;
                }

                bool issame(bool a, bool b){
                    return a == b;
                }

                bool issame(int a, int b){
                    return a == b;
                }

                bool issame(vector<int> a,vector<int>b){
                    if (a.size()!=b.size()) return false;
                    for (int i=0;i<a.size();i++)
                    {
                        if (a[i]!=b[i]) return false;
                    }
                    return true;
                }

                bool issame(string a, string b){
                    return a == b;
                }

                bool issame(vector<float> a,vector<float>b){
                    if (a.size()!=b.size()) return false;
                    for (int i=0;i<a.size();i++)
                    {
                        if (abs(a[i]-b[i])>1e-4) return false;
                    }
                    return true;
                }

                bool issame(double a, double b){
                    return abs(a - b) < 1e-3;
                }

                bool issame(vector<vector<int>> a,vector<vector<int>> b){
                    if (a.size()!=b.size()) return false;
                    for (int i=0;i<a.size();i++)
                    {
                        if (a[i].size()!=b[i].size()) return false;
                        for (int j=0;j<a[i].size();j++)
                        {
                            if (a[i][j]!=b[i][j]) return false;
                        }
                    }
                    return true;
                }

                bool issame(map<char,int> a,map<char,int> b){
                    if (a.size()!=b.size()) return false;
                    map <char,int>::iterator it;
                    for (it=a.begin();it!=a.end();it++)
                    {
                        char w1=it->first;
                        int w2=it->second;
                        if (b.find(w1)==b.end()) return false;
                        if (b[w1]!=w2) return false;
                    }
                    return true;
                }

                bool issame(long long a, long long b){
                    return a == b;
                }
            '''
            code_gen = textwrap.dedent(prefix) + code_gen

            # print out the result of the equality check to console, use double brackets to override f string defaults
            ending = f'''\
                int main() {{
                    cout << issame({entry_point}({test_input}), {test_output});
                    return 0;
                }}
            '''
            code_gen = code_gen + '\n' + textwrap.dedent(ending)
            with open(f'test_code_{rank}.cpp', 'w') as f:
                f.write(code_gen)
            compilation_process = subprocess.run(
                [
                    'g++',
                    '-std=c++11',
                    f'test_code_{rank}.cpp',
                    '-o',
                    f'test_code_{rank}',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            if compilation_process.returncode == 0:
                run_process = subprocess.run(f'./test_code_{rank}', stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                output = run_process.stdout.decode()
            else:
                output = '0'

            if f'test_code_{rank}.cpp' in os.listdir():
                os.remove(f'test_code_{rank}.cpp')
            if f'test_code_{rank}' in os.listdir():
                os.remove(f'test_code_{rank}')
            val.value = output == '1'
        elif language == 'javascript':
            ending = '\nconsole.log(JSON.stringify(' + entry_point + '(' + test_input + ')) === JSON.stringify(' + test_output + '));'
            code_gen = code_gen + ending

            with open(f'test_code_{rank}.js', 'w') as f:
                f.write(code_gen)

            run_process = subprocess.run(
                ['node', f'test_code_{rank}.js'],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            output = run_process.stdout.decode()
            if f'test_code_{rank}.js' in os.listdir():
                os.remove(f'test_code_{rank}.js')
            val.value = output == 'true\n'

        elif language == 'c':
            # since we can't run tostring based comparisons in C, initialize a series of overloaded equality functions
            prefix = '''\
                #include <stdio.h>
                #include <stdbool.h>
                #include <math.h>
                #include <stdlib.h>

                bool issame_int(int a, int b){
                    return a == b;
                }

                bool issame_bool(bool a, bool b){
                    return a == b;
                }

                bool issame_float(float a, float b){
                    return fabs(a-b) < 1e-4;
                }

                #define issame(a, b) _Generic((a), int: issame_int, bool: issame_bool, float: issame_float)(a, b)
            '''
            code_gen = textwrap.dedent(prefix) + code_gen

            # print out the result of the equality check to console, use double brackets to override f string defaults
            ending = f'''\
                int main() {{
                    bool val = issame({entry_point}({test_input}), {test_output});
                    printf("%d", val);
                    return 0;
                }}
            '''
            code_gen = code_gen + '\n' + textwrap.dedent(ending)
            with open(f'test_code_{rank}.c', 'w') as f:
                f.write(code_gen)
            compilation_process = subprocess.run(
                ['gcc', f'test_code_{rank}.c', '-o', f'test_code_{rank}'],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            if compilation_process.returncode == 0:
                run_process = subprocess.run(f'./test_code_{rank}', stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                output = run_process.stdout.decode()
            else:
                output = '0'

            if f'test_code_{rank}.c' in os.listdir():
                os.remove(f'test_code_{rank}.c')
            if f'test_code_{rank}' in os.listdir():
                os.remove(f'test_code_{rank}')
            val.value = output == '1'
        else:
            raise ValueError(f'Language {language} not supported.')



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""AWS Lambda compatible eval client."""
import json
import logging
import os

import requests

from composer.utils.eval_client.eval_client import EvalClient

__all__ = ['LambdaEvalClient']
log = logging.getLogger(__name__)


class LambdaEvalClient(EvalClient):
    """Utility for creating a client for and invoking an AWS Lambda."""

    def __init__(self) -> None:
        """Checks that the requisite environment variables are in the EvalClient.

        There must be CODE_EVAL_URL for the URL of the lambda API and CODE_EVAL_APIKEY
        for the API key of the lambda API.
        """
        if 'CODE_EVAL_URL' not in os.environ or 'CODE_EVAL_APIKEY' not in os.environ:
            raise Exception(
                'Please set environment variable CODE_EVAL_URL to the URL of the lambda API '
                'and CODE_EVAL_APIKEY to the API key of the API gateway.',
            )
        log.debug('Running code eval on lambda.')

    def invoke(self, payload: list[list[list[dict[str, str]]]]) -> list[list[list[bool]]]:
        """Invoke a batch of provided payloads for code evaluations."""
        ret = []
        for prompt_group in payload:
            ret_prompt_group = []
            for generation_group in prompt_group:
                ret_generation_group = []
                for test_case in generation_group:
                    ret_generation_group.append(self.invoke_helper(test_case))
                ret_prompt_group.append(ret_generation_group)
            ret.append(ret_prompt_group)
        return ret

    def invoke_helper(self, payload: dict[str, str]) -> bool:
        """Invoke a provided dictionary payload to the client."""
        response = requests.post(
            os.environ['CODE_EVAL_URL'],
            data=json.dumps(payload),
            headers={'x-api-key': os.environ['CODE_EVAL_APIKEY']},
        )
        response = response.json()
        return 'statusCode' in response and response['statusCode'] == 200



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Profiler Marker."""

from __future__ import annotations

import functools
import time
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence, Type, Union

from composer.profiler.trace_handler import TraceHandler

if TYPE_CHECKING:
    from composer.core import State, Timestamp

__all__ = ['Marker']


class Marker:
    """Profiler Marker.

    Used by the :class:`.Engine` to measure the duration of :class:`.Event` during training.

    .. note::

        :class:`.Marker` should not be instantiated directly; instead use :meth:`.Profiler.marker`.

    Markers can record the following types of events:

    #. Duration: Records the start and stop time of an event of interest (:meth:`.Marker.start()`, :meth:`.Marker.finish()`).
    #. Instant: Record time a particular event occurs, but not the full duration (:meth:`.Marker.instant()`).
    #. Counter: The value of a variable at given time (:meth:`.Marker.counter()`).

    A :class:`.Marker` can also be used as a context manager or decorator to record a duration:

    #. Use a :class:`.Marker` with a context manager:

        .. testsetup::

            from composer.profiler import Profiler, cyclic_schedule
            profiler = Profiler(schedule=cyclic_schedule(), trace_handlers=[], torch_prof_memory_filename=None)
            profiler.bind_to_state(state)

        .. doctest::

            >>> def something_to_measure():
            ...     print("something_to_measure")
            >>> marker = profiler.marker("foo")
            >>> with marker:
            ...     something_to_measure()
            something_to_measure

    #. Use a :class:`.Marker` as a decorator:

        .. testsetup::

            from composer.profiler import Profiler, cyclic_schedule
            profiler = Profiler(schedule=cyclic_schedule(), trace_handlers=[], torch_prof_memory_filename=None)
            profiler.bind_to_state(state)

        .. doctest::

            >>> marker = profiler.marker("foo")
            >>> @marker
            ... def something_to_measure():
            ...     print("something_to_measure")
            >>> something_to_measure()
            something_to_measure
    """

    def __init__(
        self,
        state: State,
        should_record: Callable[[State], bool],
        trace_handlers: Sequence[TraceHandler],
        name: str,
        record_instant_on_start: bool,
        record_instant_on_finish: bool,
        categories: Union[list[str], tuple[str, ...]],
    ) -> None:
        self.state = state
        self.trace_handlers = trace_handlers
        self.name = name
        self.categories = categories
        self.record_instant_on_start = record_instant_on_start
        self.record_instant_on_finish = record_instant_on_finish
        self.should_record = should_record
        self._started = False
        self._recorded_start = False

    def _record_duration_event(self, is_start: bool, wall_clock_time_ns: int, timestamp: Timestamp):
        """Record a duration event."""
        for handler in self.trace_handlers:
            handler.process_duration_event(
                name=self.name,
                categories=self.categories,
                timestamp=timestamp,
                is_start=is_start,
                wall_clock_time_ns=wall_clock_time_ns,
            )

    def _record_instant_event(self, wall_clock_time_ns: int, timestamp: Timestamp):
        """Record an instant event."""
        for handler in self.trace_handlers:
            handler.process_instant_event(
                name=self.name,
                categories=self.categories,
                timestamp=timestamp,
                wall_clock_time_ns=wall_clock_time_ns,
            )

    def _record_counter_event(
        self,
        wall_clock_time_ns: int,
        timestamp: Timestamp,
        values: dict[str, Union[int, float]],
    ) -> None:
        """Record a counter invent."""
        for handler in self.trace_handlers:
            handler.process_counter_event(
                name=self.name,
                categories=self.categories,
                wall_clock_time_ns=wall_clock_time_ns,
                timestamp=timestamp,
                values=values,
            )

    def start(self) -> None:
        """Record the start of a duration event.

        To record the duration of an event, invoke :meth:`.Marker.start` followed by :meth:`.Marker.finish`.

        .. testsetup::

            from composer.profiler import Profiler, cyclic_schedule
            profiler = Profiler(schedule=cyclic_schedule(), trace_handlers=[], torch_prof_memory_filename=None)
            profiler.bind_to_state(state)

        .. doctest::

            >>> def something_to_measure():
            ...     print("something_to_measure")
            >>> marker = profiler.marker("foo")
            >>> marker.start()
            >>> something_to_measure()
            something_to_measure
            >>> marker.finish()
        """
        if self._started:
            raise RuntimeError(
                f'Attempted to start profiler event {self.name}; however, this marker is already started',
            )

        self._recorded_start = self.should_record(self.state)
        if self._recorded_start:
            wall_clock_time = time.time_ns()
            self._record_duration_event(
                is_start=True,
                wall_clock_time_ns=wall_clock_time,
                timestamp=self.state.timestamp,
            )
            if self.record_instant_on_start:
                self._record_instant_event(
                    timestamp=self.state.timestamp,
                    wall_clock_time_ns=wall_clock_time,
                )
        self._started = True

    def finish(self) -> None:
        """Record the end of a duration event.

        See :meth:`.Marker.start()` for a usage example.
        """
        if not self._started:
            raise RuntimeError(
                f'Attempted to finish profiler event {self.name}; however, this profiler event is not yet started',
            )

        wall_clock_time = time.time_ns()
        self._record_duration_event(
            is_start=False,
            timestamp=self.state.timestamp,
            wall_clock_time_ns=wall_clock_time,
        )
        if self.record_instant_on_finish:
            self._record_instant_event(
                wall_clock_time_ns=wall_clock_time,
                timestamp=self.state.timestamp,
            )

        self._started = False

    def instant(self) -> None:
        """Record an instant event.

        To record an instant event:

        .. testsetup::

            from composer.profiler import Profiler, cyclic_schedule
            profiler = Profiler(schedule=cyclic_schedule(), trace_handlers=[], torch_prof_memory_filename=None)
            profiler.bind_to_state(state)

        .. doctest::

            >>> def something_to_measure():
            ...     print("something_to_measure")
            >>> marker = profiler.marker("instant")
            >>> marker.instant()
            >>> something_to_measure()
            something_to_measure
        """
        if self.should_record(self.state):
            self._record_instant_event(
                wall_clock_time_ns=time.time_ns(),
                timestamp=self.state.timestamp,
            )

    def counter(self, values: dict[str, Union[float, int]]) -> None:
        """Record a counter event.

        To record a counter event:

        .. testsetup::

            from composer.profiler import Profiler, cyclic_schedule
            profiler = Profiler(schedule=cyclic_schedule(), trace_handlers=[], torch_prof_memory_filename=None)
            profiler.bind_to_state(state)

        .. doctest::

            >>> marker = profiler.marker("foo")
            >>> counter_event = 5
            >>> marker.counter({"counter_event": counter_event})
            >>> counter_event = 10
            >>> marker.counter({"counter_event": counter_event})
        """
        if self.should_record(self.state):
            self._record_counter_event(
                wall_clock_time_ns=time.time_ns(),
                values=values,
                timestamp=self.state.timestamp,
            )

    def __enter__(self) -> Marker:
        self.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        del exc_type, exc, traceback  # unused
        self.finish()

    def __call__(self, func: Optional[Callable[..., Any]] = None) -> Callable[..., Any]:
        if func is None:
            # for decorators of the style @Marker(),
            # return self so it's equivalent to @Marker
            return self

        @functools.wraps(func)
        def wrapped(*args: Any, **kwargs: Any):
            with self:
                func(*args, **kwargs)

        return wrapped
